# -*- coding: utf-8 -*-
#
#  Copyright (C) 2017, 2020, 2023-2024 Rocky Bernstein
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import os.path as osp
from typing import Optional

import pyficache

from trepan.lib.stack import frame2file
from trepan.processor.parse.semantics import Location

INVALID_LOCATION = None


def resolve_location(proc, location) -> Optional[Location]:
    """Expand fields in Location namedtuple. If:
    '.':  get fields from stack
    function/module: get fields from evaluation/introspection
    location file and line number: use that
    """
    curframe = proc.curframe
    offset = None
    if location == ".":
        if not curframe:
            proc.errmsg("Don't have a stack to get location from")
            return INVALID_LOCATION
        filename = frame2file(proc.core, curframe, canonic=False)
        lineno = inspect.getlineno(curframe)
        offset = curframe.f_lasti
        return Location(filename, lineno, False, curframe.f_code, offset)

    assert isinstance(location, Location)
    is_address = location.is_address
    if proc.curframe:
        g = curframe.f_globals
        locals_dict = curframe.f_locals
    else:
        g = globals()
        locals_dict = locals()
        pass

    location_method = location.method
    filename = lineno = mod_or_func_or_code = None
    if location_method:
        # Validate arguments that can't be done in parsing
        if location_method == "<module>":
            filename = location.path
            if not filename:
                proc.errmsg("Can only resolve <module> if a path is given")
                return INVALID_LOCATION

            if osp.exists(filename):
                # from sys.modules, pick out those modules whose filename is "module_path".
                modules = [
                    module
                    for module in sys.modules.values()
                    if hasattr(module, "__file__") and module.__file__ == filename
                ]
                if len(modules):
                    # There is at least one matching module. (They all
                    # should be the same.)
                    mod_or_func_or_code = modules[0]
                else:
                    proc.errmsg(f"Cannot resolve <module> from a path {filename}")
                    return INVALID_LOCATION

        if mod_or_func_or_code is None:
            # [1] DRY similar code [2] below
            msg = f"Object {location_method} is not known yet as a function."
            try:
                mod_or_func_or_code = eval(location_method, g, locals_dict)
            except Exception:
                proc.errmsg(msg)
                split_names = location_method.split(".")
                if len(split_names) > 1:
                    proc.msg(f'Try importing {".".join(split_names[:-1])}?')

                return INVALID_LOCATION

            try:
                # Check if the converted string is a function or instance
                # method.  We don't want to test on attributes and not use
                # `inspect.isfunction()` so that we can accommodate
                # trepan-xpy() which has it's own type of compatible
                # Function, that would fail an `inspect.isfunction()`
                # test.
                if hasattr(mod_or_func_or_code, "__code__") or hasattr(mod_or_func_or_code, "im_func"):
                    offset = -1
                else:
                    proc.errmsg(msg)
                    return INVALID_LOCATION
            except Exception:
                proc.errmsg(msg)
                return INVALID_LOCATION
            filename = proc.core.canonic(mod_or_func_or_code.__code__.co_filename)

            # FIXME: we may want to check lineno and
            # respect that in the future
            lineno = mod_or_func_or_code.__code__.co_firstlineno

    elif location.path:
        filename = proc.core.canonic(location.path)
        lineno = location.line_number
        mod_or_func_or_code = None
        msg = f"{location.path} is not known as a file"
        if not osp.isfile(filename):
            # See if argument is a module
            try:
                mod_or_func_or_code = eval(location.path, g, locals_dict)
            except Exception:
                msg = f"Don't see '{location.path}' as a existing file or as an module"
                proc.errmsg(msg)
                return INVALID_LOCATION
            pass
            is_address = location.is_address
            if inspect.ismodule(mod_or_func_or_code):
                if hasattr(mod_or_func_or_code, "__file__"):
                    filename = pyficache.resolve_name_to_path(mod_or_func_or_code.__file__)
                    filename = proc.core.canonic(filename)
                    if not lineno:
                        # use first line of module file
                        lineno = 1
                        offset = 0
                        is_address = False
                    return Location(filename, lineno, is_address, mod_or_func_or_code, offset)
                else:
                    msg = f"module '{location.path}' doesn't have a file associated with it"

            proc.errmsg(msg)
            return INVALID_LOCATION
        maxline = pyficache.maxline(filename)
        if maxline and lineno > maxline:
            # NB: we use the gdb wording here
            proc.errmsg(
                "Line number %d out of range; %s has %d lines."
                % (lineno, filename, maxline)
            )
            return INVALID_LOCATION
        offset = location.offset
        if offset is None:
            code_info, lineinfo = pyficache.code_line_info(filename, lineno)
            if lineinfo:
                offset = lineinfo[0].offsets[0]
                mod_or_func_or_code = code_info[lineinfo[0].name]
            else:
                return INVALID_LOCATION


    elif location.line_number:
        if curframe is None:
            proc.errmsg("Current frame is not set")
            return INVALID_LOCATION
        filename = frame2file(proc.core, curframe, canonic=False)
        lineno = location.line_number
        is_address = location.is_address
        mod_or_func_or_code = curframe.f_code
        if offset is None:
            code_info, lineinfo = pyficache.code_line_info(filename, lineno, include_offsets=True)
            if lineinfo:
                offset = lineinfo[0].offsets[0]
                mod_or_func_or_code_name = lineinfo[0].name
                if mod_or_func_or_code.co_name != mod_or_func_or_code_name:
                    # Breakpoint is in a nested function/method.
                    # Get new code object
                    mod_or_func_or_code = code_info.get(mod_or_func_or_code_name, mod_or_func_or_code)
                pass
            else:
                return INVALID_LOCATION
    elif location.offset is not None:
        filename = frame2file(proc.core, curframe, canonic=False)
        is_address = True
        lineno = None
        mod_or_func_or_code = None
        offset = location.offset
    return Location(filename, lineno, is_address, mod_or_func_or_code, offset)


def resolve_address_location(proc, location) -> Optional[Location]:
    """Expand fields in Location namedtuple. If:
    '.':  get fields from stack
    function/module: get fields from evaluation/introspection
    location file and line number: use that
    """
    curframe = proc.curframe
    if location == ".":
        filename = frame2file(proc.core, curframe, canonic=False)
        offset = curframe.f_lasti
        is_address = True
        return Location(filename, offset, False, None, offset)

    assert isinstance(location, Location)
    is_address = True
    if proc.curframe:
        g = curframe.f_globals
        locals_dict = curframe.f_locals
    else:
        g = globals()
        locals_dict = locals()
        pass
    if location.method:
        # Validate arguments that can't be done in parsing
        filename = offset = None
        msg = f"Object {location.method} is not known yet as a function."
        # [2] DRY simlar code above [1]
        try:
            mod_or_func_or_code = eval(location.method, g, locals_dict)
        except Exception:
            split_names = location.method.split(".")
            if len(split_names) > 1:
                proc.msg(f'Try importing {".".join(split_names[:-1])}?')
            proc.errmsg(msg)
            return INVALID_LOCATION

        try:
            # Check if the converted string is a function or instance method
            if inspect.isfunction(mod_or_func_or_code) or hasattr(mod_or_func_or_code, "im_func"):
                pass
            else:
                proc.errmsg(msg)
                return INVALID_LOCATION
        except Exception:
            proc.errmsg(msg)
            return INVALID_LOCATION
        filename = proc.core.canonic(mod_or_func_or_code.func_code.co_filename)
        # FIXME: we may want to check offset and
        # respect that in the future
        offset = 0
    elif location.path:
        filename = proc.core.canonic(location.path)
        offset = location.line_number
        is_address = location.is_address
        mod_or_func_or_code = None
        msg = f"{location.path} is not known as a file"
        if not osp.isfile(filename):
            # See if argument is a module
            try:
                mod_or_func_or_code = eval(location.path, g, locals_dict)
            except Exception:
                msg = f"Don't see '{location.path}' as a existing file or as an module"
                proc.errmsg(msg)
                return INVALID_LOCATION
            pass
            is_address = location.is_address
            if inspect.ismodule(mod_or_func_or_code):
                if hasattr(mod_or_func_or_code, "__file__"):
                    filename = pyficache.resolve_name_to_path(mod_or_func_or_code.__file__)
                    filename = proc.core.canonic(filename)
                    if not offset:
                        # use first offset of module file
                        offset = 0
                        is_address = True
                    return Location(filename, offset, is_address, mod_or_func_or_code, offset)
                else:
                    msg = f"module '{location.path}' doesn't have a file associated with it"

            proc.errmsg(msg)
            return INVALID_LOCATION
        maxline = pyficache.maxline(filename)
        if maxline and offset > maxline:
            # NB: we use the gdb wording here
            proc.errmsg(
                f"Line number {offset} out of range; {filename} has {maxline} lines."
            )
            return INVALID_LOCATION
    elif location.line_number is not None:
        filename = frame2file(proc.core, curframe, canonic=False)
        offset = location.line_number
        is_address = location.is_address
        mod_or_func_or_code = proc.list_object
    else:
        proc.errmsg(
            f"Location {location} doesn't have enough information for a location."
        )
        return INVALID_LOCATION

    return Location(filename, offset, is_address, mod_or_func_or_code, offset)


# Demo it
if __name__ == "__main__":
    import sys

    from trepan.processor.cmdproc import CommandProcessor
    from trepan.processor.command.mock import MockDebugger

    d = MockDebugger()
    cmdproc = CommandProcessor(d.core)
    frame = cmdproc.frame = sys._getframe()
    cmdproc.setup()
    location = Location(__file__, frame.f_lineno, False, None, frame.f_lasti)
    print(resolve_location(cmdproc, location))
    location = Location(__file__, None, False, "resolve_location", frame.f_lasti)
    print(resolve_location(cmdproc, location))
    location = Location(__file__, None, False, "<module>", frame.f_lasti)
    print(resolve_location(cmdproc, location))



"""
Functional test of debugger "skip" command.
"""

from os.path import basename
from pathlib import Path
from test.functional.fn_helper import compare_output, strarray_setup

# FIXME: try this:
import pyficache

absolute_path = str(Path(__file__).absolute())
short_name = basename(__file__)
pyficache.update_cache(short_name)
pyficache.file2file_remap.update({short_name: absolute_path})


def test_skip():
    # See that we can skip without parameter. (Same as 'skip 1'.)
    cmds = ["step", "skip", "continue"]
    d = strarray_setup(cmds)
    d.core.start()
    ##############################
    x = 3
    x = 4
    y = 5
    y = 7  # NOQA
    ##############################
    d.core.stop()
    # x = 4 is shown in prompt, but not *run*.
    out = [
        "-- x = 3",
        "-- x = 4",
        "-- y = 5",
        "-- y = 7  # NOQA"
    ]
    compare_output(out, d)
    assert x == 3, "should have skipped x=4"

    # See that we can skip with a count value
    cmds = ["step", "skip 2", "continue"]
    d = strarray_setup(cmds)
    d.core.start()
    ##############################
    x = 7
    x = 8
    x = 9
    y = 10  # NOQA
    ##############################
    d.core.stop(options={"remove": True})
    out = [
        "-- x = 7",
        "-- x = 8",
        "-- y = 10  # NOQA",
        '-- d.core.stop(options={"remove": True})',
    ]  # x = 8 is shown in prompt, but not run.
    compare_output(out, d)
    assert x == 7, "x = 8..9 should have been skipped"
    return



# -*- coding: utf-8 -*-
#   Copyright (C) 2015, 2017, 2020, 2024-2025 Rocky Bernstein <rocky@gnu.org>
"""Breakpoints as used in a debugger.

This code is a rewrite of the stock python bdb.Breakpoint"""

__all__ = ["BreakpointManager", "Breakpoint"]

import os.path as osp
from collections import defaultdict
from types import CodeType, ModuleType
from typing import Optional
from types import FrameType
from xdis import load_module


class Breakpoint:
    """Breakpoint class implements temporary breakpoints, ignore
    counts, disabling and (re)-enabling breakpoints and breakpoint
    conditionals.

    When matching an offset, if offset is None, the offset value is not
    considered in a match. If offset is -1, we are at a "call" event;
    the first offset in bytecode is 0.
    """

    def __init__(
        self,
        number: int,
        filename: str,
        line: int,
        temporary=False,
        condition=None,
        code=None,
        offset: Optional[int] =None,
    ):
        self.offset = offset
        self.condition = condition
        self.enabled = True

        self.filename = filename
        if filename:
            self.filename = osp.realpath(filename)

        # Needed if funcname is not None.
        self.func_first_executable_line = None
        self.code = code

        # Number of time breakpoint has been hit
        self.hits = 0

        # Number of times to ignore breakpoint before stopping
        self.ignore = 0

        self.line = line
        self.number = number

        # Delete breakpoint after hitting it.
        self.temporary = temporary
        return

    def __str__(self):
        if self.temporary:
            disp = "del  "
        else:
            disp = "keep "
        if self.enabled:
            disp = disp + "yes  "
        else:
            disp = disp + "no   "
        if self.offset is None:
            offset_str = " any"
        else:
            offset_str = "%4d" % self.offset
        if self.offset is None:
            offset_str = " any"
        else:
            offset_str = "%4d" % self.offset
        if self.line is None:
            line_str = ""
        else:
            line_str = ":%d" % self.line

        msg = "%-4dbreakpoint   %s %s at %s%s" % (
            self.number,
            disp,
            offset_str,
            self.filename,
            line_str,
        )
        if self.condition:
            msg += f"\n\tstop only if {self.condition}"
        if self.ignore:
            msg += f"\n\tignore next {self.ignore} hits"
        if self.hits:
            ss = "" if self.hits > 1 else "s"
            msg += f"\n\tbreakpoint already hit {self.hits} time{ss}"
        return msg

    def enable(self):
        self.enabled = True
        return self.enabled

    def disable(self):
        self.enabled = False
        return self.enabled

    def icon_char(self) -> str:
        """Return a one-character "icon" giving the state of the breakpoint
        't': temporary breakpoint
        'B': enabled breakpoint
        'b': disabled breakpoint
        """
        if self.temporary:
            return "t"
        elif self.enabled:
            return "B"
        return "b"

    pass  # end of Breakpoint class


class BreakpointManager:
    """Manages the list of Breakpoints.

    Breakpoints are indexed by number in the `bpbynumber' list, and
    through a (file,line) tuple which is a key in the `bplist'
    dictionary. If the breakpoint is a function it is in `code_list' as
    well.  Note there may be more than one breakpoint per line which
    may have different conditions associated with them.
    """

    def __init__(self):

        # self.reset()

        # The below duplicates self.reset(). However we include it here,
        # to assist linters which as of 2014 largely do not grok attributes of
        # class unless it is put inside __init__

        self.bpbynumber: list = [None]
        self.bplist = defaultdict(list)
        self.code_list = defaultdict(list)
        return

    def bpnumbers(self):
        """Returns a list of strings of breakpoint numbers"""
        return ["%d" % bp.number for bp in self.bpbynumber if bp is not None]

    def get_breakpoint(self, i) -> tuple:
        if isinstance(i, str):
            try:
                i = int(i)
            except ValueError:
                return (False, f"Breakpoint value {i!r} is not a number.", None)
            pass
        if 1 == len(self.bpbynumber):
            return (False, "No breakpoints set.", None)
        elif i >= len(self.bpbynumber) or i <= 0:
            return (
                False,
                "Breakpoint number %d out of range 1..%d."
                % (i, len(self.bpbynumber) - 1),
                None,
            )
        bp = self.bpbynumber[i]
        if bp is None:
            return (False, "Breakpoint %d previously deleted." % i, None)
        return (True, None, bp)

    def add_breakpoint(
        self,
        filename: Optional[str],
        lineno: Optional[int] = None,
        offset: int = -1,
        temporary: bool = False,
        condition: Optional[str] = None,
        func_or_code=None,
    ):
        """
        Add a breakpoint in ``filename`` at line number ``lineno``.
        If ``offset`` is given and not -1, then it we must also be at that offset in order to stop.
        ``temporary`` specifies whether the breakpoint will be removed once it is hit.
        `condition`` specifies that a string Python expression to be evaluated to determine
        whether the breakpoint is hit or not.
        """
        bpnum = len(self.bpbynumber)
        if filename:
            filename = osp.realpath(filename)

        assert (
            isinstance(filename, str) or func_or_code is not None
        ), "You must either supply a filename or give a line number"

        if isinstance(func_or_code, CodeType):
            code = func_or_code
        elif isinstance(func_or_code, ModuleType):
            if hasattr(func_or_code, "__cached__"):
                # FIXME: we can probably do better hooking into importlib
                # or something lower-level
                _, _, _, code, _, _, _ = load_module(
                    func_or_code.__cached__, fast_load=True, get_code=True
                )
            else:
                print(f"Don't know what to do with frozen module {func_or_code}")
                return
        elif hasattr(func_or_code, "__code__"):
            code = func_or_code.__code__
        elif hasattr(func_or_code, "f_code"):
            code = func_or_code.f_code
        else:
            print(f"Don't know what to do with {func_or_code}, {type(func_or_code)}")
            return
        brkpt = Breakpoint(bpnum, filename, lineno, temporary, condition, code, offset)

        # Build the internal lists of breakpoints
        self.bpbynumber.append(brkpt)
        self.bplist[filename, lineno].append(brkpt)
        if func_or_code and offset in [None, -1]:
            self.code_list[code].append(brkpt)
        return brkpt

    def delete_all_breakpoints(self) -> str:
        """
        Delete all breakpoints. Return a message indicating
        what was deleted.
        """
        bp_list = []
        for bp in self.bpbynumber:
            if bp:
                bp_list.append(str(bp.number))
                self.delete_breakpoint(bp)
                pass
        if not bp_list:
            return "There are no breakpoints"
        return f"Deleted breakpoints {', '.join(bp_list)}"

    def delete_breakpoint(self, bp: Breakpoint) -> bool:
        "remove breakpoint `bp'"
        bpnum = bp.number
        self.bpbynumber[bpnum] = None  # No longer in list
        index = (bp.filename, bp.line)
        if index not in self.bplist:
            return False
        self.bplist[index].remove(bp)
        if not self.bplist[index]:
            # No more breakpoints for this file:line combo
            del self.bplist[index]
        return True

    def delete_breakpoint_by_number(self, bpnum: int) -> tuple:
        "Remove a breakpoint given its breakpoint number."
        success, msg, bp = self.get_breakpoint(bpnum)
        if not success:
            return False, msg
        self.delete_breakpoint(bp)
        return (True, "")

    def en_disable_all_breakpoints(self, do_enable=True):
        "Enable or disable all breakpoints."
        bp_list = [bp for bp in self.bpbynumber if bp]
        bp_nums = []
        if do_enable:
            endis = "en"
        else:
            endis = "dis"
            pass
        if not bp_list:
            return f"No breakpoints to {endis}able"
        for bp in bp_list:
            bp.enabled = do_enable
            bp_nums.append(str(bp.number))
            pass
        return f"Breakpoints {endis}abled: {', '.join(bp_nums)}"

    def en_disable_breakpoint_by_number(self, bpnum: int, do_enable=True) -> tuple:
        "Enable or disable a breakpoint given its breakpoint number."
        success, msg, bp = self.get_breakpoint(bpnum)
        if not success:
            return success, msg
        if do_enable:
            endis = "en"
        else:
            endis = "dis"
            pass
        if bp.enabled == do_enable:
            return (
                False,
                (f"Breakpoint ({str(bpnum)!r}) previously {endis}abled"),
            )
        bp.enabled = do_enable
        return (True, "")

    def delete_breakpoints_by_lineno(self, filename: str, lineno: int):
        """Removes all breakpoints at a give filename and line number.
        Returns a list of breakpoints numbers deleted.
        """
        if (filename, lineno) not in self.bplist:
            return []
        breakpoints = self.bplist[(filename, lineno)]
        bpnums = [bp.number for bp in breakpoints]
        for bp in list(breakpoints):
            self.delete_breakpoint(bp)
        return bpnums

    def find_bp(self, filename: str, lineno: int, frame):
        """Determine which breakpoint for this file:line is to be acted upon.

        Called only if we know there is a bpt at this
        location.  Returns breakpoint that was triggered and a flag
        that indicates if it is ok to delete a temporary breakpoint.

        """
        possibles = self.bplist[filename, lineno]
        for i in range(0, len(possibles)):
            b = possibles[i]
            if not b.enabled:
                continue
            if not checkfuncname(b, frame):
                continue
            # Count every hit when bp is enabled
            b.hits += 1
            if not b.condition:
                # If unconditional, and ignoring, go on to next, else
                # break
                if b.ignore > 0:
                    b.ignore = b.ignore - 1
                    continue
                else:
                    # breakpoint and marker that's ok to delete if
                    # temporary
                    return (b, True)
            else:
                # Conditional bp.
                # Ignore count applies only to those bpt hits where the
                # condition evaluates to true.
                try:
                    val = eval(b.condition, frame.f_globals, frame.f_locals)
                    if val:
                        if b.ignore > 0:
                            b.ignore = b.ignore - 1
                            # continue
                        else:
                            return (b, True)
                    # else:
                    #   continue
                except Exception:
                    # if eval fails, most conservative thing is to
                    # stop on breakpoint regardless of ignore count.
                    # Don't delete temporary, as another hint to user.
                    return (b, False)
                pass
            pass
        return (None, None)

    def last(self):
        return len(self.bpbynumber) - 1

    def reset(self):
        """A list of breakpoints by breakpoint number.  Each entry is
        None or an instance of Breakpoint.  Index 0 is unused, except
        for marking an effective break .... see effective()."""
        self.bpbynumber = [None]

        # A list of breakpoints indexed by (file, lineno) tuple
        self.bplist = {}
        self.code_list = {}

        return

    pass  # BreakpointManager


def checkfuncname(brkpt: Breakpoint, frame: FrameType):
    """
    Check whether we should break at `frame` because the frame's
    code object matches `brkpt.code`.
    """
    if not brkpt.code:
        # Breakpoint was set via line number.
        if brkpt.line != frame.f_lineno:
            # Breakpoint was set at a line with a def statement and the function
            # defined is called: don't break.
            return False
        return True

    # Breakpoint set via function code object

    if frame.f_code != brkpt.code:
        # It's not a function call, but rather execution of def statement.
        return False

    # We are in the right frame.
    if not brkpt.func_first_executable_line:
        # The function is entered for the 1st time.
        brkpt.func_first_executable_line = frame.f_lineno

    if brkpt.func_first_executable_line != frame.f_lineno:
        # But we are not at the first line number: don't break.
        return False
    return True


# Demo

if __name__ == "__main__":

    def foo(bp, bpmgr):
        frame = inspect.currentframe()
        assert frame
        print(f"Stop at bp2: {checkfuncname(bp, frame)}")
        # frame.f_lineno is constantly updated. So adjust for the
        # line difference between the add_breakpoint and the check.
        bp3 = bpmgr.add_breakpoint(__file__, 0, frame.f_lineno + 1, func_or_code=foo)
        print(f"Stop at bp3: {checkfuncname(bp3, frame)}")
        return

    def bar() -> int:
        frame = inspect.currentframe()
        assert frame is not None
        return frame.f_lasti

    bpmgr = BreakpointManager()
    print(bpmgr.last())
    line_number = foo.__code__.co_firstlineno
    bp = bpmgr.add_breakpoint(__file__, line_number, func_or_code=foo)
    assert bp
    print(bp.icon_char())
    print(bpmgr.last())
    print(repr(bp))
    print(str(bp))
    bp.disable()
    print(str(bp))
    import xdis

    bp = bpmgr.add_breakpoint(xdis.__file__, offset=0, func_or_code=xdis)
    print(bp)
    for i in 10, 1:
        status, msg = bpmgr.delete_breakpoint_by_number(i)
        print(f"Delete breakpoint {i}: {status} {msg}")
    import inspect

    frame = inspect.currentframe()
    assert bp
    print(f"Stop at bp: {checkfuncname(bp, frame)}")

    bp2 = bpmgr.add_breakpoint(None, None, -1, False, None, foo)
    foo(bp2, bpmgr)
    bp3 = bpmgr.add_breakpoint(
        __file__, line_number, 0, temporary=True, func_or_code=foo
    )
    assert bp3
    print(bp3.icon_char())
    print(bpmgr.bpnumbers())

    line_number = bar.__code__.co_firstlineno
    bp = bpmgr.add_breakpoint(__file__, line_number, 0, func_or_code=bar)
    assert bp
    filename = bp.filename
    assert filename
    line_number = bar()
    for i in range(3):
        bp = bpmgr.add_breakpoint(None, None, line_number, func_or_code=bar)
    print(bpmgr.delete_breakpoints_by_lineno(filename, line_number))
    print(bpmgr.delete_breakpoints_by_lineno(filename, line_number))
    print(bpmgr.bpnumbers())



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013-2014, 2016-2017, 2025 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger Server Input/Output interface. """

import socket, errno

from typing import Final
from trepan.lib import default as Mdefault
from trepan import misc as Mmisc
from trepan.inout import tcpfns as Mtcpfns
from trepan.inout.base import DebuggerInOutBase


# FIXME: Consider using Python's socketserver/SocketServer?
class TCPServer(DebuggerInOutBase):
    """Debugger Server Input/Output Socket."""

    DEFAULT_INIT_OPTS: Final = {"open": True, "socket": None}

    def __init__(self, inout=None, opts=None):
        get_option = lambda key: Mmisc.option_set(opts, key, self.DEFAULT_INIT_OPTS)

        self.inout = None
        self.conn = None
        self.addr = None
        self.remote_addr = ""
        self.buf = ""  # Read buffer
        self.line_edit = False  # Our name for GNU readline capability
        self.state = "disconnected"
        self.PORT = None
        self.HOST = None
        if inout:
            self.inout = inout
        if get_option("socket"):
            self.inout = opts["socket"]
            self.inout.listen(1)
            self.state = "listening"
        elif get_option("open"):
            self.open(opts)
            pass
        return

    def close(self):
        """Closes both socket and server connection."""
        self.state = "closing"
        if self.inout:
            self.inout.close()
            pass
        self.state = "closing connection"
        if self.conn:
            self.conn.close()
        self.state = "disconnected"
        return

    def open(self, opts=None):
        get_option = lambda key: Mmisc.option_set(
            opts, key, Mdefault.SERVER_SOCKET_OPTS
        )

        self.HOST = get_option("HOST")
        self.PORT = get_option("PORT")
        self.reuse = get_option("reuse")
        self.search_limit = get_option("search_limit")
        self.inout = None

        this_port = self.PORT - 1
        for i in range(self.search_limit):
            this_port += 1
            for res in socket.getaddrinfo(
                self.HOST,
                this_port,
                socket.AF_UNSPEC,
                socket.SOCK_STREAM,
                0,
                socket.AI_PASSIVE,
            ):
                af, socktype, proto, canonname, sa = res
                try:
                    self.inout = socket.socket(af, socktype, proto)
                except socket.error:
                    self.inout = None
                    continue
                try:
                    if get_option("reuse"):
                        # The following socket option allows the OS to reclaim
                        # The address faster on termination.
                        self.inout.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

                        pass
                    self.inout.bind(sa)
                    self.inout.listen(1)
                    self.state = "listening"
                    break
                except socket.error as exc:
                    if exc.errno in [errno.EADDRINUSE, errno.EINVAL]:
                        self.inout.close()
                        self.inout = None
                        continue
                    raise
                pass
            if self.state == "listening":
                break
        if self.inout is None:
            raise IOError(
                "could not open server socket after trying ports "
                "%s..%s" % (self.PORT, this_port)
            )
        self.PORT = this_port
        return

    def read(self):
        if len(self.buf) == 0:
            self.read_msg()
            pass
        return self.buf

    def read_msg(self):
        """Read one message unit. It's possible however that
        more than one message will be set in a receive, so we will
        have to buffer that for the next read.
        EOFError will be raised on EOF.
        """
        if self.state != "connected":
            self.wait_for_connect()
            pass
        if self.state == "connected":
            if 0 == len(self.buf):
                self.buf = self.conn.recv(Mtcpfns.TCP_MAX_PACKET)
                if 0 == len(self.buf):
                    self.state = "disconnected"
                    raise EOFError
                pass
            self.buf, data = Mtcpfns.unpack_msg(self.buf)
            return data.decode("utf-8")
        else:
            raise IOError("read_msg called in state: %s." % self.state)

    def wait_for_connect(self):
        self.conn, self.addr = self.inout.accept()
        self.remote_addr = ":".join(str(v) for v in self.addr)
        self.state = "connected"
        return

    def write(self, msg):
        """This method the debugger uses to write. In contrast to
        writeline, no newline is added to the end to `str'. Also
        msg doesn't have to be a string.
        """
        if self.state != "connected":
            self.wait_for_connect()
            pass
        buffer = Mtcpfns.pack_msg(msg)
        while len(buffer) > Mtcpfns.TCP_MAX_PACKET:
            self.conn.send(buffer[: Mtcpfns.TCP_MAX_PACKET])
            buffer = buffer[Mtcpfns.TCP_MAX_PACKET :]
        return self.conn.send(buffer)


# Demo
if __name__ == "__main__":
    inout = TCPServer(opts={"open": False})
    import sys

    if len(sys.argv) > 1:
        inout.open()
        print("Listening for connection on %s:%s" % (inout.HOST, inout.PORT))
        while True:
            try:
                line = inout.read_msg().rstrip("\n")
                print(line)
                inout.writeline("ack: " + line)
            except EOFError:
                break
            pass
        pass
    inout.close()
    pass



"Integration test of invoking debugger without a file name"
import sys

import pytest
from helper import run_debugger


@pytest.mark.skipif(
    sys.platform == "win32", reason="Need to adjust for filesystem on MS Windows"
)
def test_no_script():
    """Test invoking debugger without a file name"""
    result = run_debugger(testname="noscript", dbgr_opts="--basename", python_file=None)
    assert result, "debugger 'step' command comparision"
    return



# -*- coding: utf-8 -*-
#   Modification of Python's Lib/dis.py
# FIXME: see if we can use more of Lib/dis in Python3
"""Disassembly Routines"""

import inspect
import sys
import types
from typing import Callable

from pyficache import highlight_string
from pygments.token import Comment
from xdis import (
    IS_PYPY,
    Bytecode,
    findlabels,
    findlinestarts,
    get_instructions_bytes,
    get_opcode,
)
from xdis.std import distb
from xdis.version_info import PYTHON_VERSION_TRIPLE

from trepan.lib.format import (  # Opcode,
    Arrow,
    Details,
    Hex,
    Integer,
    Keyword,
    LineNumber,
    Number,
    Symbol,
    format_token,
)

_have_code = (types.MethodType, types.FunctionType, types.CodeType, type)


def _try_compile(source, name):
    """Attempts to compile the given source, first as an expression and
    then as a statement if the first approach fails.

    Utility function to accept strings in functions that otherwise
    expect code objects
    """
    try:
        c = compile(source, name, "eval")
    except SyntaxError:
        c = compile(source, name, "exec")
    return c


# Modified from dis. Changed output to use msg, msg_nocr, section, and
# pygments.  Added first_line and last_line parameters


def dis(
    msg: Callable,
    msg_nocr: Callable,
    section,
    errmsg,
    x=None,
    start_line=-1,
    end_line=None,
    relative_pos=False,
    style="none",
    start_offset=0,
    end_offset=None,
    include_header=False,
    asm_format="extended",
):
    """Disassemble classes, methods, functions, or code.

    With no argument, disassemble the last traceback.

    """
    lasti = -1
    if x is None:
        distb()
        return None, None
    if start_offset is None:
        start_offset = 0
    mess = ""
    if start_line > 1:
        mess += "from line %d " % start_line
    elif start_offset > 1:
        mess = "from offset %d " % start_offset
    if end_line:
        mess += "to line %d" % end_line
    elif end_offset:
        mess += "to offset %d" % end_offset

    sectioned = False

    # Try to dogpaddle to the code object for the type setting x
    if hasattr(types, "InstanceType") and isinstance(x, types.InstanceType):
        x = x.__class__
    if inspect.ismethod(x):
        section(f"Disassembly of {x}: {mess}")
        sectioned = True
        x = x.__func__.__code__
    elif inspect.isfunction(x) or inspect.isgeneratorfunction(x):
        section(f"Disassembly of {x}: {mess}")
        x = x.__code__
        sectioned = True
    elif inspect.isgenerator(x):
        section(f"Disassembly of {x}: {mess}")
        frame = x.gi_frame
        lasti = frame.f_last_i
        x = x.gi_code
        sectioned = True
    elif inspect.isframe(x):
        section(f"Disassembly of {x}: {mess}")
        sectioned = True
        if hasattr(x, "f_lasti"):
            lasti = x.f_lasti
            if lasti == -1:
                lasti = 0
            pass
        opc = get_opcode(PYTHON_VERSION_TRIPLE, IS_PYPY)
        x = x.f_code
        if include_header:
            header_lines = Bytecode(x, opc).info().split("\n")
            header = "\n".join([format_token(Comment, h) for h in header_lines])
            msg(header)
        pass
    elif inspect.iscode(x):
        pass

    if hasattr(x, "__dict__"):  # Class or module
        items = sorted(x.__dict__.items())
        for name, x1 in items:
            if isinstance(x1, _have_code):
                if not sectioned:
                    section(f"Disassembly of {x}: ")
                try:
                    dis(
                        msg,
                        msg_nocr,
                        section,
                        errmsg,
                        x1,
                        start_line=start_line,
                        end_line=end_line,
                        relative_pos=relative_pos,
                        asm_format=asm_format,
                    )
                    msg("")
                except TypeError:
                    _, msg, _ = sys.exc_info()
                    errmsg("Sorry:", msg)
                    pass
                pass
            pass
        pass
    elif hasattr(x, "co_code"):  # Code object
        if not sectioned:
            section(f"Disassembly of {x}: ")
        return disassemble(
            msg,
            msg_nocr,
            x,
            lasti=lasti,
            start_line=start_line,
            end_line=end_line,
            relative_pos=relative_pos,
            style=style,
            start_offset=start_offset,
            end_offset=end_offset,
            asm_format=asm_format,
        )
    elif isinstance(x, str):  # Source code
        return disassemble_string(
            msg,
            msg_nocr,
            x,
        )
    else:
        errmsg(f"Don't know how to disassemble {type(x).__name__} objects.")
    return None, None


def disassemble(
    msg: Callable,
    msg_nocr: Callable,
    co,
    lasti: int = -1,
    start_line: int = -1,
    end_line=None,
    relative_pos=False,
    style="none",
    start_offset=0,
    end_offset=None,
    asm_format="extended",
):
    """Disassemble a code object."""
    return disassemble_bytes(
        msg,
        msg_nocr,
        co.co_code,
        lasti,
        co.co_firstlineno,
        start_line,
        end_line,
        relative_pos,
        co.co_varnames,
        co.co_names,
        co.co_consts,
        co.co_cellvars,
        co.co_freevars,
        dict(findlinestarts(co)),
        style,
        start_offset=start_offset,
        end_offset=end_offset,
        opc=opc,
        asm_format=asm_format,
    )


def disassemble_string(msg, msg_nocr, source):
    """Compile the source string, then disassemble the code object."""
    return disassemble_bytes(msg, msg_nocr, _try_compile(source, "<dis>"))


opc = get_opcode(PYTHON_VERSION_TRIPLE, IS_PYPY)


def print_instruction(
    instr,
    instructions: list,
    msg: Callable,
    msg_nocr: Callable,
    labels: list,
    lasti: int=-1,
    line_starts = {},
    style="none",
    opc=opc,
    asm_format="extended",
):
    """Disassemble bytecode at offset `offsets`."""

    # Python 3.11 introduces "CACHE" and the convention seems to be
    # to not print these normally.
    if instr.opname == "CACHE" and asm_format not in (
        "extended_bytes",
        "bytes",
    ):
        return

    offset = instr.offset
    opcode = instr.opcode

    # Column: Source code line number
    if instr.starts_line:
        if offset:
            msg("")
        cur_line = instr.starts_line
        msg_nocr(format_token(LineNumber, "%4d" % cur_line, style=style))
        msg_nocr(" ")
    else:
        msg_nocr("     ")

    # Column: Current instruction indicator
    if offset == lasti:
        msg_nocr(format_token(Arrow, "-->", style=style))
    else:
        msg_nocr("   ")

    # Column: Jump target marker
    if offset in labels:
        msg_nocr(format_token(Arrow, ">>", style=style))
    else:
        msg_nocr("  ")

    # Column: Instruction offset from start of code sequence
    msg_nocr(format_token(Number, repr(offset).rjust(4), style=style))
    msg_nocr(" ")

    # Column: Instruction bytes
    if asm_format in ("extended-bytes", "bytes"):
        msg_nocr(format_token(Symbol, "|", style=style))
        msg_nocr(format_token(Hex, f"{instr.opcode:02x}", style=style))
        if instr.inst_size == 1:
            # Not 3.6 or later
            msg_nocr(" " * (2 * 3))
        if instr.inst_size == 2:
            # Must by Python 3.6 or later
            msg_nocr(" ")
            if instr.has_arg:
                msg_nocr(format_token(Hex, f"{instr.arg % 256:02x}", style=style))
            else:
                msg_nocr(format_token(Hex, "00", style=style))
        elif instr.inst_size == 3:
            # Not 3.6 or later
            opbyte, operand_byte = divmod(instr.arg, 256)
            msg_nocr(format_token(Hex, f"{opbyte:02x}", style=style))
            msg_nocr(" ")
            msg_nocr(format_token(Hex, f"{operand_byte:02x}", style=style))

        msg_nocr(format_token(Symbol, "|", style=style))
        msg_nocr(" ")

    # Column: Opcode name
    msg_nocr(format_token(Keyword, instr.opname.ljust(20), style=style))
    msg_nocr(" ")

    # Column: Opcode argument
    argrepr = None
    if instr.arg is not None:
        argrepr = instr.argrepr
        # The argrepr value when the instruction was created generally has all the information we require.
        if asm_format in ("extended", "extended-bytes"):
            if (
                instr.is_jump()
                and line_starts is not None
                and line_starts.get(instr.argval) is not None
            ):
                msg(
                    format_token(
                        Details,
                        f"{instr.argrepr}, line {line_starts[instr.argval]}",
                        style=style,
                    )
                )
                return
            if (
                hasattr(opc, "opcode_extended_fmt")
                and opc.opname[opcode] in opc.opcode_extended_fmt
            ):
                tos_str, start_offset = opc.opcode_extended_fmt[opc.opname[opcode]](
                    opc, list(reversed(instructions))
                )
                if start_offset is not None:
                    if style is not None and style != "none":
                        tos_str = highlight_string(tos_str, style=style)
                    msg(tos_str)
                    return
                pass
            pass
    elif asm_format in ("extended", "extended-bytes"):
        # Note: instr.arg is also None
        if (
            hasattr(opc, "opcode_extended_fmt")
            and opc.opname[opcode] in opc.opcode_extended_fmt
        ):
            tos_str, start_offset = opc.opcode_extended_fmt[opc.opname[opcode]](
                opc, list(reversed(instructions))
            )
            if start_offset is not None:
                msg(highlight_string(tos_str, style=style))
                return
    if argrepr is None or argrepr == "":
        if instr.arg is not None:
            msg(format_token(Integer, str(instr.arg), style=style))
        else:
            msg("")
            pass
    else:
        msg(format_token(Details, instr.argrepr, style=style))

    pass


def disassemble_bytes(
    orig_msg: Callable,
    orig_msg_nocr: Callable,
    code,
    lasti=-1,
    cur_line=0,
    start_line=-1,
    end_line=None,
    relative_pos=False,
    varnames=(),
    names=(),
    constants=(),
    cells=(),
    freevars=(),
    line_starts={},
    style="none",
    start_offset=0,
    end_offset=None,
    opc=opc,
    asm_format="extended",
) -> tuple:
    """Disassemble byte string of code. If end_line is negative
    it counts the number of statement line starts to use."""

    def null_print(_):
        return None

    instructions = []
    if end_line is None:
        end_line = 10000
    elif relative_pos:
        end_line += start_line - 1
        pass

    labels = findlabels(code, opc)

    if start_line > cur_line:
        msg_nocr = null_print
        msg = null_print
    else:
        msg_nocr = orig_msg_nocr
        msg = orig_msg

    offset = -1
    for instr in get_instructions_bytes(
        code, opc, varnames, names, constants, cells, line_starts, labels
    ):
        instructions.append(instr)

        offset = instr.offset

        if end_offset and offset > end_offset:
            break

        # Python 3.11 introduces "CACHE" and the convention seems to be
        # to not print these normally.
        if instr.opname == "CACHE" and asm_format not in (
            "extended_bytes",
            "bytes",
        ):
            continue

        # Column: Source code line number
        if instr.starts_line:
            cur_line = instr.starts_line
            if start_line and (
                (start_line > cur_line) or start_offset and start_offset > offset
            ):
                continue
            else:
                msg_nocr = orig_msg_nocr
                msg = orig_msg

            if (cur_line > end_line) or (end_offset and offset > end_offset):
                break
        elif start_offset and offset:
            if offset >= start_offset:
                msg_nocr = orig_msg_nocr
                msg = orig_msg
            else:
                continue

        print_instruction(
            instr,
            instructions,
            msg,
            msg_nocr,
            labels=labels,
            lasti=lasti,
            line_starts=line_starts,
            style=style,
            opc=opc,
            asm_format=asm_format,
        )

    return code, offset

def disassemble_instruction(
    orig_msg: Callable,
    orig_msg_nocr: Callable,
    code,
    offset,
    lasti=-1,
    cur_line=0,
    end_line=None,
    relative_pos=False,
    varnames=(),
    names=(),
    constants=(),
    cells=(),
    freevars=(),
    line_starts={},
    style="none",
    opc=opc,
    asm_format="extended",
) -> tuple:
    """Disassemble byte string of code. If end_line is negative
    it counts the number of statement line starts to use."""

    def null_print(_):
        return None

    instructions = []
    if end_line is None:
        end_line = 10000
    elif relative_pos:
        end_line += start_line - 1
        pass

    labels = findlabels(code, opc)

    if start_line > cur_line:
        msg_nocr = null_print
        msg = null_print
    else:
        msg_nocr = orig_msg_nocr
        msg = orig_msg

    offset = -1
    for instr in get_instructions_bytes(
        code, opc, varnames, names, constants, cells, line_starts, labels
    ):
        instructions.append(instr)

        offset = instr.offset

        if end_offset and offset > end_offset:
            break

        # Python 3.11 introduces "CACHE" and the convention seems to be
        # to not print these normally.
        if instr.opname == "CACHE" and asm_format not in (
            "extended_bytes",
            "bytes",
        ):
            continue

        # Column: Source code line number
        if instr.starts_line:
            cur_line = instr.starts_line
            if start_line and (
                (start_line > cur_line) or start_offset and start_offset > offset
            ):
                continue

            if (cur_line > end_line) or (end_offset and offset > end_offset):
                break
        elif start_offset and offset and start_offset <= offset:
            continue

        print_instruction(
            instr,
            instructions,
            msg,
            msg_nocr,
            labels=labels,
            lasti=lasti,
            line_starts=line_starts,
            style=style,
            opc=opc,
            asm_format=asm_format,
        )

    return code, offset

# Demo it
if __name__ == "__main__":

    def msg(msg_str):
        print(msg_str)
        return

    def msg_nocr(msg_str):
        sys.stdout.write(msg_str)
        return

    def errmsg(msg_str):
        msg("*** " + msg_str)
        return

    def section(msg_str):
        msg("=== " + msg_str + " ===")
        return

    def fib(x):
        if x <= 1:
            return 1
        return fib(x - 1) + fib(x - 2)

    curframe = inspect.currentframe()
    # dis(msg, msg_nocr, errmsg, section, curframe,
    #     start_line=10, end_line=40, highlight='dark')
    # print('-' * 40)
    # does nothing because start_offset is too high:
    # dis(msg, msg_nocr, errmsg, section, curframe,
    #     start_offset=10, end_offset=20, highlight='dark')
    print("-" * 40)
    # for asm_format in ("std", "extended", "bytes", "extended-bytes"):
    for asm_format in ("extended", "bytes", "extended-bytes"):
        print("Format is", asm_format)
        dis(
            msg,
            msg_nocr,
            section,
            errmsg,
            disassemble,
            asm_format=asm_format,
            style="tango",
        )
        dis(msg, msg_nocr, section, errmsg, fib, asm_format=asm_format, style="tango")
        print("=" * 30)

    # print('-' * 40)
    # magic, moddate, modtime, co = pyc2code(sys.modules['types'].__file__)
    # disassemble(msg, msg_nocr, section, co, -1, 1, 70)
    pass



"""Unit test for "trepan.frame" """
import inspect
import pytest
import platform

from trepan.lib.stack import count_frames, is_eval_or_exec_stmt


def test_count_frames():
    f = inspect.currentframe()
    frame_count = count_frames(f)
    assert count_frames(f) > 2
    assert frame_count - 1 == count_frames(f.f_back)
    assert frame_count - 1 == count_frames(f, 1)
    return


@pytest.mark.skipif(platform.python_implementation() == "GraalVM",
                    reason="exec/eval detection doesn't work for Graal (JVM) bytecode")
def test_stack_misc():
    f = inspect.currentframe()
    # assert "test_stack_misc" == get_call_function_name(f))
    assert is_eval_or_exec_stmt(f) is None
    exec("result = is_eval_or_exec_stmt(inspect.currentframe())")
    assert "eval" == eval(" is_eval_or_exec_stmt(inspect.currentframe())")
    return



"""
Functional test of debugger "step" command.
"""

import os
from os.path import basename
from pathlib import Path
from test.functional.fn_helper import compare_output, strarray_setup

import pyficache
import pytest
import tracer
from xdis import PYTHON_VERSION_TRIPLE

absolute_path = str(Path(__file__).absolute())
short_name = basename(__file__)
pyficache.update_cache(short_name)
pyficache.file2file_remap.update({short_name: absolute_path})


def test_step_same_level():
    # See that we can step with parameter which is the same as 'step 1'
    cmds = ["step", "continue"]
    d = strarray_setup(cmds)
    d.core.start()
    ##############################
    x = 5  # NOQA
    y = 6  # NOQA
    ##############################
    d.core.stop()
    out = ["-- x = 5  # NOQA", "-- y = 6  # NOQA"]
    compare_output(out, d)
    return


def test_step_computed_value():
    # See that we can step with a computed count value
    cmds = ["step 5-3", "continue"]
    d = strarray_setup(cmds)
    d.core.start()
    ##############################
    x = 5
    y = 6
    z = 7
    ##############################
    d.core.stop(options={"remove": True})
    out = ["-- x = 5", "-- z = 7"]
    compare_output(out, d)

    # Test step>
    cmds = ["step>", "continue"]
    d = strarray_setup(cmds)
    d.core.start()
    ##############################
    x = 5

    def foo():
        return

    y = 6  # NOQA
    foo()
    ##############################
    d.core.stop(options={"remove": True})
    out = ["-- x = 5", "-> def foo():"]
    compare_output(out, d)

    # Test step!
    cmds = ["step!", "continue"]
    d = strarray_setup(cmds)
    d.core.start()
    ##############################
    x = 5
    try:
        y = 2
        z = 1 / 0
    except:
        pass
    ##############################
    d.core.stop(options={"remove": True})
    out = ["-- x = 5", "!! z = 1 / 0"]
    compare_output(out, d)

    # Test "step" with sets of events. Part 1
    cmds = ["step call exception", "step call exception", "continue"]
    d = strarray_setup(cmds)

    # d.core.start()
    ##############################
    # x = 5  # NOQA
    # try:

    #     def foo1():
    #         y = 2  # NOQA
    #         raise Exception
    #         return

    #     foo1()
    # except:
    #     pass
    # z = 1  # NOQA
    # ##############################
    # d.core.stop(options={'remove': True})
    # out = ['-- x = 5  # NOQA',
    #        '-> def foo1():',
    #        '!! raise Exception']
    # compare_output(out, d)

    # # Test "step" will sets of events. Part 2
    # cmds = ['step call exception 1+0',
    #         'step call exception 1', 'continue']
    # d = strarray_setup(cmds)
    # d.core.start()
    # ##############################
    # x = 5
    # try:
    #     def foo2():
    #         y = 2
    #         raise Exception
    #         return
    #     foo2()
    # except:
    #     pass
    # z = 1
    # ##############################
    # d.core.stop(options={'remove': True})
    # out = ['-- x = 5',
    #        '-> def foo2():',
    #        '!! raise Exception']
    # compare_output(out, d)

    return


@pytest.mark.skipif(
    "CI" in os.environ, reason="Need to figure out what's up on CircleCI"
)
def test_step_between_fn():
    # Step into and out of a function
    def sqr(x):
        return x * x

    if PYTHON_VERSION_TRIPLE < (3, 10):
        test2_expect = [
            "-- d.core.start()",
            "-- x = sqr(4)  # NOQA",
            "-> def sqr(x):",
            "-- return x * x",
            "<- return x * x",
        ]
    else:
        test2_expect = [
            "-- x = sqr(4)  # NOQA",
            "-> def sqr(x):",
            "-- return x * x",
            "<- return x * x",
            "-- y = 5  # NOQA",
        ]

    for cmds, out, eventset in (
        (
            ["step", "step", "continue"],
            [
                "-- x = sqr(4)  # NOQA",
                "-- return x * x",
                "-- y = 5  # NOQA",
            ],
            frozenset(("line",)),
        ),
        (
            ["step", "step", "step", "step", "continue"],
            test2_expect,
            tracer.ALL_EVENTS,
        ),
    ):
        d = strarray_setup(cmds)
        d.settings["events"] = eventset
        d.core.start()
        ##############################
        x = sqr(4)  # NOQA
        y = 5  # NOQA
        ##############################
        d.core.stop(options={"remove": True})
        compare_output(out, d)
        pass
    return


def test_step_in_exception():
    return

    def boom(x):
        y = 0 / x  # NOQA
        return

    def bad(x):
        boom(x)
        return x * x

    cmds = [
        "step",
        "step",
        "step",
        "step",
        "step",
        "step",
        "step",
        "step",
        "step",
        "step",
        "continue",
    ]
    d = strarray_setup(cmds)
    with pytest.raises(ZeroDivisionError):
        d.core.start()
        x = bad(0)
        assert False, "should have raised an exception"

    d.core.stop(options={"remove": True})

    out = [
        "-- x = bad(0)  # NOQA",  # line event
        "-> def bad(x):",  # call event
        "-- boom(x)",  # line event
        "-> def boom(x):",  # call event
        "-- y = 0/x  # NOQA",  # line event
        "!! y = 0/x  # NOQA",  # exception event
        "<- y = 0/x  # NOQA",  # return event
        "!! boom(x)",  # exception event
        "<- boom(x)",  # return event
        "!! x = bad(0)  # NOQA",  # exception event
        "-- except ZeroDivisionError:",
    ]
    compare_output(out, d)
    return



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2009, 2012-2013, 2020, 2023-2025 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Bytecode instruction routines"""

import re

from opcode import opname
from types import FrameType
from typing import Optional
from xdis import PYTHON_VERSION_TRIPLE, get_opcode_module

opcode_module = get_opcode_module(PYTHON_VERSION_TRIPLE)


def opname_at_code_offset(bytecode_bytes: bytes, offset: int) -> str:
    try:
        opcode = bytecode_bytes[offset]
    except IndexError:
        return "got IndexError"
    return opname[opcode]



def op_at_frame(frame: FrameType, offset: Optional[int]=None, skip_cache=True):
    bytecode = frame.f_code.co_code
    if offset is None:
        offset = frame.f_lasti
    while skip_cache and offset != 0 and opname_at_code_offset(bytecode, offset) == "CACHE":
        offset -= 2
    return opname_at_code_offset(bytecode, offset)


def next_opcode(code, offset):
    """Return the next opcode and offset as a tuple. Tuple (-100,
    -1000) is returned when reaching the end."""
    n = len(code)
    while offset < n:
        op = code[offset]
        yield op, offset
        offset = offset + 2
        yield op, offset
        pass
    yield -100, -1000
    pass


def next_linestart(co, offset: int, count=1) -> int:
    code = co.co_code

    linestarts = dict(opcode_module.findlinestarts(co))
    n = len(code)
    # contains_cond_jump = False
    while offset < len(code):
        if offset in linestarts:
            count -= 1
            if 0 == count:
                return linestarts[offset]
            pass
        offset += 2

    return -1000


def stmt_contains_opcode(co, lineno, query_opname) -> bool:
    linestarts = dict(opcode_module.findlinestarts(co))
    code = co.co_code
    found_start = False
    offset = 0
    start_line = None
    for offset, start_line in list(linestarts.items()):
        if start_line == lineno:
            print("=" * 30)
            found_start = True
            break
        pass
    if not found_start:
        return False
    for op, offset in next_opcode(code, offset):
        linestart = linestarts.get(offset)
        if -1000 == offset or linestart and linestart != start_line:
            return False
        op_name = opname[op]
        # debug: print opname
        if query_opname == op_name:
            return True
        pass
    return False


_re_def_str = r"^\s*def\s"
_re_def = re.compile(_re_def_str)


def is_def_stmt(line: Optional[str], frame: FrameType) -> bool:
    """Return True if we are looking at a def statement"""
    # Should really also check that operand of 'LOAD_CONST' is a code object
    return bool(
        line is not None
        and _re_def.match(line)
        and op_at_frame(frame) == "LOAD_CONST"
        and stmt_contains_opcode(frame.f_code, frame.f_lineno, "MAKE_FUNCTION")
    )


_re_class = re.compile(r"^\s*class\s")


def is_class_def(line, frame):
    """Return True if we are looking at a class definition statement"""
    return (
        line
        and _re_class.match(line)
        and stmt_contains_opcode(frame.f_code, frame.f_lineno, "BUILD_CLASS")
    )


# Demo stuff above
if __name__ == "__main__":
    import inspect

    def sqr(x):
        return x * x

    frame = inspect.currentframe()
    assert frame
    co = frame.f_code
    lineno = frame.f_lineno
    print(
        "contains MAKE_FUNCTION %s"
        % stmt_contains_opcode(co, lineno - 4, "MAKE_FUNCTION")
    )
    print(
        f"contains MAKE_FUNCTION {stmt_contains_opcode(co, lineno, 'MAKE_FUNCTION')}"
    )

    print(f"op at frame: {op_at_frame(frame)}")
    print(f"op at frame, position 2: {op_at_frame(frame, 2)}")
    print(f"def statement: x=5?: {is_def_stmt('x=5', frame)}")
    # Not a "def" statement because frame is wrong spot
    print(is_def_stmt("def foo():", frame))

    class Foo:
        pass

    lineno = frame.f_lineno
    print(
        f"contains BUILD_CLASS {stmt_contains_opcode(co, lineno - 2, 'BUILD_CLASS')}"
    )
    print(f"contains BUILD_CLASS {stmt_contains_opcode(co, lineno, 'BUILD_CLASS')}")
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2009-2010, 2013-2015, 2017, 2023-2024 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger input possibly attached to a user or interactive. """

import io
import os
import os.path as osp
import sys

from trepan.clifns import default_configfile
from trepan.inout import base as Mbase

try:
    from prompt_toolkit import HTML, PromptSession
    from prompt_toolkit.enums import EditingMode
    from prompt_toolkit.history import FileHistory
    from prompt_toolkit.styles import Style
except:
    PromptSession = lambda history: None
    FileHistory = lambda history: None
    HTML = lambda string: string
else:
    from trepan.inout.ptk_bindkeys import bindings, read_inputrc, read_init_file

    USER_INPUTRC = os.environ.get("TREPAN3K_INPUTRC", default_configfile("inputrc"))

    read_inputrc(read_init_file, use_unicode=False)
    if osp.isfile(USER_INPUTRC):
        if os.access(USER_INPUTRC, os.R_OK):
            read_init_file(USER_INPUTRC)
        else:
            sys.stderr.write(f"Can't read user inputrc file {USER_INPUTRC}; skipping\n")


class DebuggerUserInput(Mbase.DebuggerInputBase):
    """Debugger input connected to what we think of as a end-user input
    as opposed to a relay mechanism to another process. Input could be
    interactive terminal, but it might be file input."""

    def __init__(self, inp=None, opts=dict()):

        if opts.get("readline") == "prompt_toolkit":

            edit_mode = opts.get("edit_mode", "emacs")
            prompt_editing_mode = (
                EditingMode.EMACS if edit_mode == "emacs" else EditingMode.VI
            )
            self.session = PromptSession(
                editing_mode=prompt_editing_mode,
                enable_history_search=True,
                history=FileHistory(opts.get("histfile")),
                key_bindings=bindings,
            )

            @bindings.add("escape", "c-j")
            def toggle_editmode(_):
                self.session.editing_mode = (
                    EditingMode.VI
                    if self.session.editing_mode == EditingMode.EMACS
                    else EditingMode.EMACS
                )
                print(f"\nedit mode is now {self.session.editing_mode}")

            self.input = self.session.input
            self.line_edit = True
            self.closed = False
        else:
            self.session = None
            self.input = inp or sys.stdin
            self.line_edit = None  # Our name for GNU readline capability
            self.open(self.input, opts)

        return

    def close(self):
        self.input.close()
        self.closed = True
        return

    def use_history(self):
        return True

    def open(self, inp, opts={}):
        """Use this to set where to read from.

        """
        if isinstance(inp, "string".__class__):  # FIXME
            inp = open(inp, "r")
        elif not (
            isinstance(inp, io.TextIOWrapper)
            or isinstance(inp, io.StringIO)
            or hasattr(inp, "isatty")
            and inp.isatty()
        ):
            raise IOError("Invalid input type (%s) for %s" % (type(inp), inp))
        self.input = inp
        self.line_edit = bool(opts and opts.get("readline"))

        self.closed = False
        return

    def readline(self, use_raw=None, prompt=""):
        """Read a line of input. EOFError will be raised on EOF.

        Note: some user interfaces may decide to arrange to call
        DebuggerOutput.write() first with the prompt rather than pass
        it here
        """
        if self.session:
            # Using prompt_toolkit
            html_prompt = HTML(f"<u>{prompt.strip()}</u> ")
            line = self.session.prompt(html_prompt, style=Style.from_dict({"": ""}))
            return line.rstrip("\n")

        try:
            inp = input(prompt)
            return inp
        except ValueError:
            raise EOFError
        return

    pass


# Demo
if __name__ == "__main__":
    inp = DebuggerUserInput(__file__)
    line = inp.readline()
    print(line)
    inp.close()
    inp.open("input.py", opts={"use_raw": False})
    while True:
        try:
            inp.readline()
        except EOFError:
            break
        pass
    try:
        inp.readline()
    except EOFError:
        print("EOF handled correctly")
    pass

    if len(sys.argv) > 1:
        inp = DebuggerUserInput()
        try:
            print("Type some characters:", end=" ")
            line = inp.readline()
            print("You typed: %s" % line)
            print("Type some more characters (raw):", end=" ")
            line = inp.readline(True)
            print("Type even more characters (not raw):", end=" ")
            line = inp.readline(True)
            print("You also typed: %s" % line)
        except EOFError:
            print("Got EOF")
        pass
    pass



# This code fragment what is embedded inside ../test_skippy
# We have it here to make it easy to test things when things go wrong.


# commands are:
#   step
#   skip
#   continue
##############################
x = 3
x = 4
y = 5
y = 7  # NOQA
##############################
# commands are:
#   step
#   skip 2
#   continue
x = 7
x = 8
x = 9
y = 10  # NOQA
##############################



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2020, 2024 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import tracer

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.cmdfns import want_different_line


class StepCommand(DebuggerCommand):
    """**step**[**+**|**-**|**<**|**>**|**!**] [*event*...] [*count*]

    Execute the current line, stopping at the next event.

    With an integer argument, step that many times.

    *event* is list of an event name which is one of: "call",
    "return", "line", "exception", "c-call`, "c-return" or "c-exception".

    If specified, only those stepping events will be considered. If no
    list of event names is given, then any event triggers a stop when the
    count is 0.

    There is however another way to specify an event: you can
    suffix one of the symbols `<`, `>`, or `!` after the command or on
    an alias of that.  A suffix of `+` on a command or an alias forces a
    move to another line, while a suffix of `-` disables this requirement.
    A suffix of `>` will continue until the next call. (`finish` will run
    run until the return for that call.)

    If no suffix is given, the debugger setting `different-line`
    determines this behavior.

    Examples:
    ---------

      step        # step 1 event, *any* event
      step 1      # same as above
      step 5/5+0  # same as above
      step line   # step only line events
      step call   # step only call events
      step>       # same as above
      step call line # Step line *and* call events

    Related and similar is the `next` command.

    See also:
    ---------

    `next`, `skip`, `jump` (there's no `hop` yet), `continue`, `return` and
    `finish` for other ways to progress execution."""

    aliases = (
        "step+",
        "step-",
        "step>",
        "step<",
        "step!",
        "s",
        "s+",
        "s-",
        "s<",
        "s>",
        "s!",
    )
    execution_set = ["Running"]
    short_help = "Step program (possibly entering called functions)"

    DebuggerCommand.setup(locals(), category="running", max_args=1, need_stack=True)

    def run(self, args):
        step_events = []
        if args[0][-1] == ">":
            step_events = ["call"]
        elif args[0][-1] == "<":
            step_events = ["return"]
        elif args[0][-1] == "!":
            step_events = ["exception"]
            pass
        if len(args) <= 1:
            self.proc.debugger.core.step_ignore = 0
        else:
            pos = 1
            while pos < len(args):
                arg = args[pos]
                if arg in tracer.ALL_EVENT_NAMES:
                    step_events.append(arg)
                else:
                    break
                pos += 1
                pass
            if pos == len(args) - 1:
                self.core.step_ignore = self.proc.get_int(
                    args[pos], default=1, cmdname="step"
                )
                if self.core.step_ignore is None:
                    return False
                # 0 means stop now or step 1, so we subtract 1.
                self.core.step_ignore -= 1
                pass
            elif pos != len(args):
                self.errmsg(f"Invalid additional parameters {' '.join(args[pos])}")
                return False
            pass

        if [] == step_events:
            self.core.step_events = None
        else:
            self.core.step_events = step_events
            pass

        self.core.different_line = want_different_line(
            args[0], self.settings["different"]
        )
        self.core.stop_level = None
        self.core.last_frame = None
        self.core.stop_on_finish = False
        self.proc.continue_running = True  # Break out of command read loop
        return True

    pass


if __name__ == "__main__":
    from mock import MockDebugger

    d = MockDebugger()
    cmd = StepCommand(d.core.processor)
    for c in (["s", "5"], ["step", "1+2"], ["s", "foo"]):
        d.core.step_ignore = 0
        cmd.proc.continue_running = False
        result = cmd.run(c)
        print(f"Execute result: {result}")
        print(f"step_ignore {repr(d.core.step_ignore)}")
        print(f"continue_running: {cmd.proc.continue_running}")
        pass
    for c in (["s"], ["step+"], ["s-"], ["s!"], ["s>"], ["s<"]):
        d.core.step_ignore = 0
        cmd.continue_running = False
        result = cmd.run(c)
        print(f"different line {c[0]}:", cmd.core.different_line)
        pass
    pass



"""
Functional test of debugger "jump" command.
"""

import inspect
from os.path import basename
from pathlib import Path
from test.functional.fn_helper import compare_output, strarray_setup

# FIXME: try this:
import pyficache

absolute_path = str(Path(__file__).absolute())
short_name = basename(__file__)
pyficache.update_cache(short_name)
pyficache.file2file_remap.update({short_name: absolute_path})


def test_jump():
    # See that we can jump with line number
    curframe = inspect.currentframe()
    cmds = ["step", "jump %d" % (curframe.f_lineno + 7), "continue"]
    d = strarray_setup(cmds)  # 1
    d.core.start()  # 2
    ##############################          # 3...
    x = 4
    x = 5
    x = 6
    z = 7  # NOQA
    ##############################
    d.core.stop(options={"remove": True})
    out = [
        "-- x = 4",  # x = 4 is shown in prompt, but not run.
        "-- x = 5",
        "-- z = 7  # NOQA",
    ]
    compare_output(out, d)
    assert x == 4, " x = 5 and 6 should have been skipped"
    return



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2017, 2020 Rocky Bernstein
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from trepan.lib.stack import frame2file
from trepan.processor.location import resolve_location
from trepan.processor.parse.parser import LocationError
from trepan.processor.parse.scanner import ScannerError
from trepan.processor.parse.semantics import Location, RangeError, build_location, build_range

INVALID_PARSE_LIST = (None, None, None)


def parse_list_cmd(proc, args, listsize=10):
    """Parses arguments for the "list" command and returns the tuple:
    (filename, first line number, last line number)
    or sets these to None if there was some problem."""

    text = proc.current_command[len(args[0]) + 1 :].strip()

    if text in frozenset(("", ".", "+", "-")):
        if text == ".":
            location = resolve_location(proc, ".")
            return location.path, location.line_number, listsize
        else:
            if proc.list_lineno is None:
                proc.errmsg("Don't have previous list location")
                return INVALID_PARSE_LIST
            filename = proc.list_filename
            if text == "+":
                first = max(1, proc.list_lineno + listsize)
            elif text == "-":
                if proc.list_lineno == 1 + listsize:
                    proc.errmsg("Already at start of %s." % proc.list_filename)
                    return INVALID_PARSE_LIST
                first = max(1, proc.list_lineno - (2 * listsize) - 1)
            elif text == "":
                # Continue from where we last left off
                first = proc.list_lineno + 1
        last = first + listsize - 1
        return filename, first, last
    else:
        try:
            list_range = build_range(text)
        except LocationError as e:
            proc.errmsg("Error in parsing list range at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_LIST
        except ScannerError as e:
            proc.errmsg("Lexical error in parsing list range at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_LIST
        except RangeError as e:
            proc.errmsg(e.errmsg)
            return INVALID_PARSE_LIST

        if list_range.first is None:
            # Last must have been given
            assert isinstance(list_range.last, Location)
            location = resolve_location(proc, list_range.last)
            if not location:
                return INVALID_PARSE_LIST
            last = location.line_number
            first = max(1, last - listsize)
            return location.path, first, last
        elif isinstance(list_range.first, int):
            first = list_range.first
            location = resolve_location(proc, list_range.last)
            if not location:
                return INVALID_PARSE_LIST
            filename = location.path
            last = location.line_number
            if last < first:
                # Treat as a count rather than an absolute location
                last = first + last
            return location.path, first, last
        else:
            # First is location. Last may be empty or a number
            assert isinstance(list_range.first, Location)
            first = list_range.first.line_number
            if list_range.first.path is None:
                if list_range.first.method is None:
                    filename = frame2file(proc.core, proc.curframe, canonic=False)
                else:
                    location = resolve_location(proc, list_range.first)
                    if location is None or location == INVALID_PARSE_LOCATION:
                        return INVALID_PARSE_LIST
                    filename = location.path
                    if first is None:
                        first = location.line_number
            else:
                filename = list_range.first.path

            last = list_range.last
            # if location.method:
            #     first -= listsize // 2
            if isinstance(last, str):
                # Is an offset +number
                assert last[0] == "+"
                last = first + int(last[1:])
            elif not last:
                last = first + listsize
            elif first is None:
                return INVALID_PARSE_LIST
            elif last < first:
                # Treat as a count rather than an absolute location
                last = first + last

            return filename, first, last
        pass
    return


INVALID_PARSE_LOCATION = (None, None)


def parse_location(proc, args):
    text = proc.current_command[len(args[0]) + 1 :].strip()

    if text in frozenset(("", ".")):
        if text == ".":
            location = resolve_location(proc, ".")
            return location.path, location.line_number
        else:
            filename = proc.list_filename
            if text == "":
                # Continue from where we last left off
                first = proc.list_lineno + 1
        return filename, first
    else:
        try:
            location = build_location(text)
        except LocationError as e:
            proc.errmsg("Error in parsing location at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_LOCATION
        except ScannerError as e:
            proc.errmsg("Lexical error in location at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_LOCATION
        return location


# Demo it
if __name__ == "__main__":
    import sys

    from trepan.processor.cmdproc import CommandProcessor
    from trepan.processor.command import mock as Mmock

    d = Mmock.MockDebugger()
    cmdproc = CommandProcessor(d.core)
    # print '-' * 10
    # print_source_line(sys.stdout.write, 100, 'source_line_test.py')
    # print '-' * 10
    cmdproc.frame = sys._getframe()
    cmdproc.setup()

    def five():
        return 5

    import os  # noqa

    for cmd in (
        # "list",
        # "list +",
        # "list -",
        # "list 15, 10",
        "list five()",
        # "list 9 , 5",
        # "list 7 ,",
        # "list '''c:\\tmp\\foo.bat''':1",
        # 'list """/Users/My Documents/foo.py""":2',
        # 'list build_range()',
        # 'list os:1 ,',
        "list os.path:9 ,",
    ):
        args = cmd.split(" ")
        cmdproc.current_command = cmd
        print(parse_list_cmd(cmdproc, args))
    pass



"Step integration tests"

import sys

import pytest
from helper import run_debugger
from xdis import PYTHON_VERSION_TRIPLE


@pytest.mark.skipif(
    sys.platform == "win32", reason="Need to adjust for filesystem on MS Windows"
)
def test_step():
    """Test stepping, set skip, set trace"""
    if PYTHON_VERSION_TRIPLE >= (3, 11):
        right_template = "%s-311.right"
    elif PYTHON_VERSION_TRIPLE >= (3, 10):
        right_template = "%s-310.right"
    elif PYTHON_VERSION_TRIPLE >= (3, 8):
        right_template = "%s-38.right"
    else:
        right_template = None
    result = run_debugger(
        testname="step",
        dbgr_opts="--basename --highlight=plain",
        python_file="gcd.py",
        right_template=right_template,
    )
    assert result, "debugger 'step' command comparision"
    return



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2009, 2013-2015, 2023, 2025 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger FIFO Input/Output interface. """

import os
import tempfile

from typing import Final
from trepan import misc as Mmisc
from trepan.inout.base import DebuggerInOutBase
from trepan.lib import default as Mdefault, file as Mfile


class FIFOClient(DebuggerInOutBase):
    """Debugger Client Input/Output Socket."""

    DEFAULT_INIT_OPTS: Final = {"open": True}

    def __init__(self, inp=None, opts=None):
        get_option = lambda key: Mmisc.option_set(
            opts, key, Mdefault.CLIENT_SOCKET_OPTS
        )
        self.state = "disconnected"
        self.flush_after_write = True
        self.input = None
        self.output = None
        self.line_edit = False  # Our name for GNU readline capability
        self.state = "disconnected"
        open_pid = get_option("open")
        if open_pid:
            self.open(open_pid)
            pass
        return

    def close(self):
        """Closes both input and output"""
        self.state = "closing"
        if self.input:
            self.input.close()
            pass
        if self.output:
            self.output.close()
            pass
        self.state = "disconnnected"
        return

    def flush(self):
        return self.output.flush()

    def open(self, pid, opts=None):

        # Not in/out are reversed from server side
        d = tempfile.gettempdir()
        self.out_name = os.path.join(d, ("trepan-%s.in" % pid))
        self.in_name = os.path.join(d, ("trepan-%s.out" % pid))
        is_readable = Mfile.readable(self.out_name)
        if not is_readable:
            if is_readable is None:
                raise IOError("output FIFO %s doesn't exist" % self.out_name)
            else:
                raise IOError("output FIFO %s is not readable" % self.out_name)
            pass
        is_readable = Mfile.readable(self.in_name)
        if not is_readable:
            if is_readable is None:
                raise IOError("input FIFO %s doesn't exist" % self.in_name)
            else:
                raise IOError("input FIFO %s is not readable" % self.out_name)
            pass
        self.state = "active"
        return

    def read_msg(self):
        """Read a line of input. EOFError will be raised on EOF.

        Note that we don't support prompting"""
        # FIXME: do we have to create and check a buffer for
        # lines?
        if self.state == "active":
            if not self.input:
                self.input = open(self.in_name, "r")
                pass
            line = self.input.readline()
            if not line:
                self.state = "disconnected"
                raise EOFError
            return line.encode("utf-8")
        else:
            raise IOError("readline called in state: %s." % self.state)
        return  # Not reached

    def write(self, msg):
        """This method the debugger uses to write. In contrast to
        writeline, no newline is added to the end to `str'.
        """
        if self.state == "active":
            if not self.output:
                self.output = open(self.out_name, "w")
                pass
            pass
        else:
            raise EOFError
        self.output.write(msg)
        if self.flush_after_write:
            self.flush()
        return


# Demo
if __name__ == "__main__":
    fifo = FIFOClient(opts={"open": False})
    import sys

    if len(sys.argv) > 1:
        print(
            "Connecting...",
        )
        fifo.open(sys.argv[1])
        print("connected.")
        while True:
            prompt = fifo.readline()
            line = input(prompt)
            if len(line) == 0:
                break
            try:
                line = fifo.writeline(line)
                print("Got: ", fifo.readline())
            except EOFError:
                break
            pass
        pass
    fifo.close()
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2009, 2013, 2023-2024 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
import linecache
import os
import os.path as osp

def default_configfile(base_filename: str) -> str:
    """Return fully expanded configuration filename location for
    base_filename directory: ~/.config/trepan3k
    """
    file_dir = osp.join(os.environ.get("HOME", "~"), ".config", "trepan3k")
    file_dir = path_expanduser_abs(file_dir)

    if not osp.isdir(file_dir):
        os.makedirs(file_dir, mode=0o755)
    return osp.join(file_dir, base_filename)

# FIXME: do a better job of this. Live parsing?
def is_ok_line_for_breakpoint(filename, lineno, errmsg_fn):
    """Check whether specified line seems to be executable.

    Return `lineno` if it is, 0 if not (e.g. a docstring, comment, blank
    line or EOF). Warning: testing is not comprehensive.
    """
    line = linecache.getline(filename, lineno)
    if not line:
        errmsg_fn("End of file")
        return False
    line = line.strip()
    # Don't allow setting breakpoint at a blank line
    if not line or (line[0] == "#") or (line[:3] == '"""') or line[:3] == "'''":
        errmsg_fn("Blank or comment")
        return False
    return True


def file2module(filename):
    """Given a file name, extract the most likely module name."""
    basename = osp.basename(filename)
    if "." in basename:
        pos = basename.rfind(".")
        return basename[:pos]
    else:
        return basename
    return None


def search_file(filename, directories, cdir):
    """Return a full pathname for filename if we can find one. path
    is a list of directories to prepend to filename. If no file is
    found we'll return None"""

    for trydir in directories:

        # Handle $cwd and $cdir
        if trydir == "$cwd":
            trydir = "."
        elif trydir == "$cdir":
            trydir = cdir

        tryfile = osp.realpath(osp.join(trydir, filename))
        if osp.isfile(tryfile):
            return tryfile
    return None


def whence_file(py_script, dirnames=None):
    """Do a shell-like path lookup for py_script and return the results.
    If we can't find anything return py_script"""
    if py_script.find(os.sep) != -1:
        # Don't search since this name has path separator components
        return py_script
    if dirnames is None:
        dirnames = os.environ["PATH"].split(os.pathsep)
    for dirname in dirnames:
        py_script_try = osp.join(dirname, py_script)
        if osp.exists(py_script_try):
            return py_script_try
    # Failure
    return py_script


def path_expanduser_abs(filename):
    return os.path.abspath(os.path.expanduser(filename))


# Demo
if __name__ == "__main__":
    import sys

    print(file2module(sys.argv[0]), sys.argv[0])
    ok = is_ok_line_for_breakpoint(__file__, 1, sys.stdout.write)
    print("\nCan stop at line 1? ", ok)
    ok = is_ok_line_for_breakpoint(__file__, 2, sys.stdout.write)
    print("\nCan stop at line 2? ", ok)
    print(path_expanduser_abs("./.trepan3k"))



"""
Functional test of debugger "next" command.
"""

from os.path import basename
from pathlib import Path
from test.functional.fn_helper import compare_output, strarray_setup

import pyficache
import pytest

# FIXME: try this:

absolute_path = str(Path(__file__).absolute())
short_name = basename(__file__)
pyficache.update_cache(short_name)
pyficache.file2file_remap.update({short_name: absolute_path})


def test_next_same_level():
    # See that we can next with parameter which is the same as 'next 1'
    cmds = ["next", "continue"]
    d = strarray_setup(cmds)
    d.core.start()
    x = 5
    y = 6
    d.core.stop()
    out = ["-- x = 5", "-- y = 6"]
    compare_output(out, d)

    # See that we can next with a computed count value
    cmds = ["next 5-3", "continue"]
    d = strarray_setup(cmds)
    d.core.start()
    x = 5  # NOQA
    y = 6  # NOQA
    z = 7  # NOQA
    d.core.stop(options={"remove": True})
    out = ["-- x = 5  # NOQA", "-- z = 7  # NOQA"]
    compare_output(out, d)
    return


def test_next_between_fn():
    # Next over a function
    def fact(x):
        if x <= 1:
            return 1
        return fact(x - 1)

    cmds = ["next", "continue"]
    d = strarray_setup(cmds)
    d.core.start()
    x = fact(4)  # NOQA
    y = 5  # NOQA
    d.core.stop(options={"remove": True})
    out = ["-- x = fact(4)  # NOQA", "-- y = 5  # NOQA"]
    compare_output(out, d)
    return


def test_next_in_exception():
    def boom(x):
        y = 0 / x  # NOQA
        return

    def buggy_fact(x):
        if x <= 1:
            return boom(0)
        return buggy_fact(x - 1)

    cmds = ["next", "continue"]
    d = strarray_setup(cmds)
    with pytest.raises(ZeroDivisionError):
        d.core.start()
        x = buggy_fact(4)  # NOQA
        y = 5  # NOQA
        assert False, "should have raised an exception"

    d.core.stop(options={"remove": True})

    out = ["-- x = buggy_fact(4)  # NOQA", "!! x = buggy_fact(4)  # NOQA"]
    compare_output(out, d)
    return



"""
Tests to see how well we do with eval/exec
"""

# Deparsing will handle this
eval_str = "1+2"
x = eval(eval_str)
print(x)

# When deparsing does not work, looking at the string to the
# eval call will work
x = eval("3+4")

exec("""y = 5;
x += y
""")
print(x)

# TODO: add an AST example.



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2009, 2013, 2015, 2020, 2023-2024 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import inspect
import os
from dis import findlinestarts

from trepan.processor.cmdproc import print_location

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class JumpCommand(DebuggerCommand):
    """**jump** *lineno*

    Set the next line that will be executed.

    There are a number of limitations on what line can be set.

    You can't jump:

      - into the body of a for loop

      - into an 'except' block from outside

      - outside or inside of a code block you are stopped

    Jumping to a previous line is one way to reexecuting code.
    Jumping to a return statement may be similar to you back to the caller
    without running more code in the current frame.

    See also:
    ---------

    `skip`, `next`, `step`, `jump`, `continue`, `return` and
    `finish` for other ways to progress execution.

    `eval` can be used to run Python code without changing the
    execution line.
    """

    aliases = ("j",)
    category = "running"
    execution_set = ["Running"]
    min_args = 1
    max_args = 1
    name = os.path.basename(__file__).split(".")[0]
    need_stack = False
    short_help = "Set the next line to be executed"

    def run(self, args):
        if not self.core.is_running():
            return False

        if self.proc.curindex + 1 != len(self.proc.stack):
            self.errmsg("You can only jump within the bottom frame")
            return False

        if self.proc.curframe.f_trace is None:
            self.errmsg(
                "Sigh - operation can't be done here; frame f_trace is not set."
            )
            return False

        lineno = self.proc.get_an_int(
            args[1], f"jump: a line number is required, got {args[1]}."
        )

        if lineno is None:
            return False

        linestarts = dict(findlinestarts(self.proc.curframe.f_code))
        if lineno not in linestarts.values():
            self.warnmsg(f"code for line {lineno} not found. Results may be unpredictable.")

        try:
            # Set to change position, update our copy of the stack,
            # and display the new position
            self.proc.curframe.f_lineno = lineno
            self.proc.stack[self.proc.curindex] = (
                self.proc.stack[self.proc.curindex][0],
                lineno,
            )
            print_location(self.proc)
        except ValueError as e:
            self.errmsg(f"jump failed: {e}")
        return False

    pass


if __name__ == "__main__":
    from trepan.processor.command import mock
    from trepan.processor.cmdproc import get_stack

    d, cp = mock.dbg_setup()
    command = JumpCommand(cp)
    print("jump when not running: ", command.run(["jump", "1"]))
    command.core.execution_status = "Running"
    cp.curframe = inspect.currentframe()
    cp.curindex = 0
    cp.stack = get_stack(cp.curframe, None, None)
    command.run(["jump", "1"])
    cp.curindex = len(cp.stack) - 1
    command.run(["jump", "1"])
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2008-2009, 2013-2017, 2019-2021, 2023-2025 Rocky
#   Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Some singleton debugger methods that can be called without first
creating a debugger object -- these methods will create a debugger object,
if necessary, first.
"""

# The following routines could be done via something like the following
# untested code:
# def _build_standalone(methodname, docstring=None):
#     def function(arg, arg2=None, globals=None, locals=None):
#         Debugger().getattr(methodname)(arg, arg2, globals, locals)
#        return
#    function.__name__ = methodname
#    function.__doc__ = docstring
#    return function
#
# But this a bit cumbersome and perhaps overkill for the 2 or so
# functions below.  It also doesn't work once we add the exception handling
# we see below. So for now, we'll live with the code duplication.

import os
import sys
import traceback

from trepan.debugger import Trepan, debugger_obj
from trepan.interfaces.server import ServerInterface
from trepan.lib.default import DEBUGGER_SETTINGS
from trepan.post_mortem import post_mortem_excepthook, uncaught_exception

DEFAULT_DEBUG_PORT = 1955


def debug(
    dbg_opts={},
    start_opts=None,
    post_mortem: bool = True,
    step_ignore: int = 1,
    level: int = 0,
):
    """
    Enter the debugger.

    Parameters
    ----------

    level : how many stack frames go back. Usually it will be
    the default 0. But sometimes though there may be calls in setup to the debugger
    that you may want to skip.

    step_ignore : how many line events to ignore after the
    debug() call. 0 means don't even wait for the debug() call to finish.

    dbg_opts: is an optional "options" dictionary that gets fed
              trepan.Debugger(); `
    start_opts: are the optional "options" dictionary that gets fed to
                trepan.Debugger.core.start().
    post_mortem: start debugger in post-mortem mode.

    Use like this:

    .. code-block:: python

        ... # Possibly some Python code
        import trepan.api # Needed only once
        ... # Possibly some more Python code
        trepan.api.debug() # You can wrap inside conditional logic too
        pass  # Stop will be here.
        # Below is code you want to use the debugger to do things.
        ....  # more Python code
        # If you get to a place in the program where you aren't going
        # want to debug anymore, but want to remove debugger trace overhead:
        trepan.api.stop()

    Parameter "level" specifies how many stack frames go back. Usually it will be
    the default 0. But sometimes though there may be calls in setup to the debugger
    that you may want to skip.

    Parameter "step_ignore" specifies how many line events to ignore after the
    debug() call. 0 means don't even wait for the debug() call to finish.

    In situations where you want an immediate stop in the "debug" call
    rather than the statement following it ("pass" above), add parameter
    step_ignore=0 to debug() like this::

        import trepan.api  # Needed only once
        # ... as before
        trepan.api.debug(step_ignore=0)
        # ... as before

    Module variable _debugger_obj_ from module ``trepan.debugger`` is used as
    the debugger instance variable; it can be subsequently used to change
    settings or alter behavior. It should be of type Debugger (found in
    module trepan). If not, it will get changed to that type::

       $ python
       >>> from trepan.debugger import debugger_obj
       >>> type(debugger_obj)
       <type 'NoneType'>
       >>> import trepan.api
       >>> trepan.api.debug()
       ...
       (Trepan) c
       >>> from trepan.debugger import debugger_obj
       >>> debugger_obj
       <trepan.debugger.Debugger instance at 0x7fbcacd514d0>
       >>>

    If however you want your own separate debugger instance, you can
    create it from the debugger _class Debugger()_ from module
    trepan.debugger::

      $ python
      >>> from trepan.debugger import Debugger
      >>> dbgr = Debugger()  # Add options as desired
      >>> dbgr
      <trepan.debugger.Debugger instance at 0x2e25320>

    `dbg_opts' is an optional "options" dictionary that gets fed
    trepan.Debugger(); `start_opts' are the optional "options"
    dictionary that gets fed to trepan.Debugger.core.start()."""

    # We have a global debugger_obj that we reuse
    global debugger_obj

    # A list of debugger profiles we might run
    dbg_initfiles = []

    if debugger_obj is None:
        # If debugger_obj has not been set this is the first time
        # we are entering the debugger.
        # create the object, and set to run the user's
        # debugger initialization profile

        debugger_obj = Trepan(dbg_opts)
        debugger_obj.core.add_ignore(debug, stop)

        # Run user profile if first time and we haven't
        # explicit set to ignore profile loading.
        if not start_opts or start_opts.get("startup-profile", True):
            from trepan.options import add_startup_file

            add_startup_file(dbg_initfiles)

        pass

    core = debugger_obj.core
    frame = sys._getframe(0 + level)
    core.set_next(frame)

    # If we've specified profile loading, add that
    if start_opts and start_opts.get("startup-profile", False):
        from trepan.options import add_startup_file

        add_startup_file(dbg_initfiles)

    for init_cmdfile in dbg_initfiles:
        core.processor.queue_startfile(init_cmdfile)

    if not core.is_started():
        core.add_ignore(debug, stop)
        core.start(start_opts)
        pass
    if post_mortem:
        debugger_on_post_mortem()
        pass
    if 0 == step_ignore:
        frame = sys._getframe(1 + level)
        core.stop_reason = "at a debug() call"
        old_trace_hook_suspend = core.trace_hook_suspend
        core.trace_hook_suspend = True
        core.processor.event_processor(frame, "line", None)
        core.trace_hook_suspend = old_trace_hook_suspend
    else:
        core.step_ignore = step_ignore - 1
        pass
    return


def debug_for_remote_access():
    """Enter the debugger in a mode that allows connection to it
    outside of the process being debugged.
    """
    connection_opts = {
        "IO": "TCP",
        "PORT": os.getenv("TREPAN3K_TCP_PORT", DEFAULT_DEBUG_PORT),
    }
    intf = ServerInterface(connection_opts=connection_opts)
    dbg_opts = {'interface': intf}
    print('Starting %s server listening on %s.' % (connection_opts["IO"], connection_opts["PORT"]),
          file=sys.stderr)
    print('Use `python3 -m trepan.client --port %s to enter the debugger.' % connection_opts["PORT"],
          file=sys.stderr)
    debug(dbg_opts=dbg_opts, step_ignore=0, level=1)

def debugger_on_post_mortem():
    """Call debugger on an exception that terminates a program"""
    sys.excepthook = post_mortem_excepthook
    return


def run_call(func, *args, debug_opts=DEBUGGER_SETTINGS, start_opts=None, **kwds):
    """Call the function (a function or method object, not a string)
    with the given arguments starting with the statement after
    the place that this appears in your program.

    When run_call() returns, it returns whatever the function call
    returned.  The debugger prompt appears as soon as the function is
    entered."""

    dbg = Trepan(opts=debug_opts)
    try:
        return dbg.run_call(func, *args, **kwds)
    except Exception:
        uncaught_exception(dbg)
        pass
    return


def run_eval(expression, debug_opts = DEBUGGER_SETTINGS, start_opts = None, globals_ = None, locals_ = None, _ = None):
    """Evaluate the expression (given as a string) under debugger
    control starting with the statement after the place that
    this appears in your program.

    This is a wrapper to Debugger.run_eval(), so see that.

    When run_eval() returns, it returns the value of the expression.
    Otherwise, this function is similar to run().
    """

    dbg = Trepan(opts=debug_opts)
    try:
        return dbg.run_eval(
            expression, start_opts=start_opts, globals_=globals_, locals_=locals_
        )
    except Exception:
        dbg.core.trace_hook_suspend = True
        traceback.print_exc()
    finally:
        dbg.core.trace_hook_suspend = False
    return


def run_exec(
    statement,
    debug_opts = DEBUGGER_SETTINGS,
    start_opts=None,
    globals_=None,
    locals_=None,
):
    """Execute the statement (given as a string) under debugger
    control starting with the statement subsequent to the place that
    this run_call appears in your program.

    This is a wrapper to Debugger.run_exec(), so see that.

    The debugger prompt appears before any code is executed;
    you can set breakpoints and type 'continue', or you can step
    through the statement using 'step' or 'next'

    The optional globals_ and locals_ arguments specify the environment
    in which the code is executed; by default the dictionary of the
    module __main__ is used."""

    dbg = Trepan(opts=debug_opts)
    try:
        return dbg.run_exec(
            statement, start_opts=start_opts, globals_=globals_, locals_=locals_
        )
    except Exception:
        uncaught_exception(dbg)
        pass
    return

def stop(opts=None):
    if isinstance(debugger_obj, Trepan):
        return debugger_obj.stop(opts)
    return None


# # Demo it
if __name__ == "__main__":
    import tracer

    def plus5(n: int) -> int:
        return n + 5

    from trepan.inout.stringarray import StringArrayInput, StringArrayOutput
    from trepan.lib.default import DEBUGGER_SETTINGS

    settings = dict(DEBUGGER_SETTINGS)
    settings.update({"trace": True, "printset": tracer.ALL_EVENTS})
    debugger_input = StringArrayInput()
    debugger_output = StringArrayOutput()
    debug_opts = {
        "step_ignore": -1,
        "settings": settings,
        "input": debugger_input,
        "output": debugger_output,
    }
    print('Issuing: run_eval("1+2")')
    run_eval("1+2", debug_opts=debug_opts)
    print(debugger_output.output)
    # print('Issuing: run_exec("x=1; y=2")')
    # run_exec("x=1; y=2", debug_opts=debug_opts)
    debugger_input = StringArrayInput(["step", "list", "continue"])
    debugger_output = StringArrayOutput()
    debug_opts = {
        "step_ignore": -1,
        "settings": settings,
        "input": debugger_input,
        "output": debugger_output,
    }
    if len(sys.argv) > 1:
        print(
            "Issuing interactive: run_call(plus5, %s)" % sys.argv[1]
        )
        run_call(plus5, int(sys.argv[1]), debug_opts=debug_opts)
    else:
        print(
            "Issuing interactive: run_call(plus5, 2)"
        )
        run_call(plus5, 2, debug_opts=debug_opts)
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2017, 2020, 2024-2025 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from trepan.processor.cmdbreak import parse_break_cmd, set_break
from trepan.processor.command.base_cmd import DebuggerCommand


class ContinueCommand(DebuggerCommand):
    """**continue** [*location*]

    Leave the debugger read-eval print loop and continue
    execution. Subsequent entry to the debugger however may occur via
    breakpoints or explicit calls, or exceptions.

    If *location* is given, a temporary breakpoint is set at that
    position before continuing.

    Examples:
    ---------

        continue          # Continue execution
        continue 5        # Continue with a one-time breakpoint at line 5
        continue basename # Go to os.path.basename if we have basename imported
        continue /usr/lib/python3.8/posixpath.py:110 # Possibly the same as
                                                     # the above using file
                                                     # and line number

    See also:
    ---------

    `step` `jump`, `next`, `finish` and `help syntax location`"""

    aliases = ("c",)
    execution_set = ["Running"]
    short_help = "Continue execution of debugged program"
    DebuggerCommand.setup(locals(), category="running", max_args=1, need_stack=True)

    def run(self, args):
        if len(args) > 1:
            # FIXME: DRY this code. Better is to hook into tbreak.
            func, filename, lineno, condition, offset = parse_break_cmd(self.proc, args)
            if not (func is None and filename is None):
                set_break(
                    self, func, filename, lineno, condition, True, args, offset=offset
                    )
            else:
                self.errmsg(f"Did not find stopping spot for: {' '.join(args[1:])}")
        self.core.step_events = None  # All events
        self.core.step_ignore = -1
        self.proc.continue_running = True  # Break out of command read loop

        # Try to remove debugger hook if no breakpoints are set.
        self.proc.fast_continue = True
        return True

    pass


if __name__ == "__main__":
    import sys

    def five():
        return 5

    from trepan.debugger import Trepan

    d = Trepan()
    cmd = ContinueCommand(d.core.processor)
    cmd.proc.frame = sys._getframe()
    line = cmd.proc.frame.f_lineno
    # Leave this line after line_str blank!
    cmd.proc.setup()

    for c in (
        ["continue", "wrong", "number", "of", "args"],
        ["c", str(line)],
        ["c", str(line + 1)],  # Invalid
        ["continue", "1"],
        ["c", "five"],
        ["c", "five()"],
    ):
        d.core.step_ignore = 0
        cmd.continue_running = False
        cmd.proc.current_command = " ".join(c)
        result = cmd.run(c)
        print(f"Run result: {result}")
        print(
            f"step_ignore {d.core.step_ignore}, continue_running: {cmd.continue_running}"
        )
        pass
    pass



#
#   Copyright (C) 2024-2025 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Keyboard input binding routines for prompt_toolkit which are
analogous to GNU Readlines' parse_and_bind().
"""

# NOTE: this same code also exists in mathicsscript. Hopefully in the future
# we will have a way to avoid the duplication.


import pathlib
import re
from prompt_toolkit.key_binding import KeyBindings
from typing import Callable

bindings = KeyBindings()

def read_inputrc(read_init_file_fn: Callable, use_unicode: bool) -> None:
    """
    Read GNU Readline style inputrc for prompt_toolkit
    """
    # GNU Readline inputrc $include's paths are relative to itself,
    # so chdir to its directory before reading the file.
    parent_dir = pathlib.Path(__file__).parent.absolute()
    inputrc = "inputrc-unicode" if use_unicode else "inputrc-no-unicode"
    try:
        read_init_file_fn(str(parent_dir / "data" / inputrc))
    except Exception:
        pass


def read_init_file(path: str):
    def check_quoted(s: str):
        return s[0:1] == '"' and s[-1:] == '"'

    def add_binding(alias_expand, replacement: str):
        def self_insert(event):
            event.current_buffer.insert_text(replacement)

        bindings.add(*alias_expand)(self_insert)

    for line_no, line in enumerate(open(path, "r").readlines()):
        line = line.strip()
        if not line or line.startswith("#"):
            continue
        fields = re.split(r"\s*: ", line)
        if len(fields) != 2:
            print(f"{line_no+1}: expecting 2 fields, got {len(fields)} in:\n{line}")
            continue
        alias, replacement = fields
        if not check_quoted(alias):
            print(f"{line_no+1}: expecting alias to be quoted, got {alias} in:\n{line}")
        alias = alias[1:-1]
        if not check_quoted(replacement):
            print(
                f"{line_no+1}: expecting replacement to be quoted, got {replacement} in:\n{line}"
            )
            continue
        replacement = replacement[1:-1]
        alias_expand = [
            c if c != "\x1b" else "escape" for c in list(alias.replace(r"\e", "\x1b"))
        ]
        add_binding(alias_expand, replacement)
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2009-2010, 2013-2015,
#   2017-2018, 2020, 2023-2025 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Interface when communicating with the user in the same process as
    the debugged program."""
import atexit
import os.path as osp
import pathlib

from os import environ

from trepan.clifns import default_configfile
from trepan.inout.input import DebuggerUserInput
from trepan.inout.output import DebuggerUserOutput
from trepan.interface import TrepanInterface

histfile = environ.get("TREPAN3KHISTFILE", default_configfile("history"))

# Create HISTFILE if it doesn't exist already
if not osp.isfile(histfile):
    pathlib.Path(histfile).touch()

# is_pypy = '__pypy__' in sys.builtin_module_names

DEFAULT_USER_SETTINGS = {
    "histfile": histfile,  # Where do we save the history?
    "histsize": 50,  # How many history items do we save
    "complete": None,  # Function which handles tab completion, or None
}

try:
    from readline import (
        parse_and_bind,
        read_history_file,
        set_completer,
        set_history_length,
        write_history_file,
    )
except ImportError:
    def write_history_file(_: str):
        return
    def parse_and_bind(_: str):
        raise RuntimeError(f"Called {__name__}() when it doesn't exist")
    read_history_file = set_completer = parse_and_bind
    have_complete = False
else:
    have_complete = False

class UserInterface(TrepanInterface):
    """Interface when communicating with the user in the same
    process as the debugged program."""

    def __init__(self, inp=None, out=None, opts={}):
        self.user_opts = DEFAULT_USER_SETTINGS.copy()
        self.user_opts.update(opts)

        atexit.register(self.finalize)
        self.interactive = True  # Or at least so we think initially
        self.input = inp or DebuggerUserInput(self.user_opts.get("input", {}), self.user_opts)
        self.output = out or DebuggerUserOutput()
        self.debugger_name = self.user_opts.get("debugger_name", "trepan3k")
        self.histfile = None

        if self.input.use_history():
            self.complete = self.user_opts["complete"] if have_complete else None
            if self.complete is not None:
                parse_and_bind("tab: complete")
                set_completer(self.complete)
                pass
            self.histfile = self.user_opts["histfile"]
            if self.histfile:
                try:
                    read_history_file(self.histfile)
                except IOError:
                    pass
                except Exception:
                    # PyPy read_history_file fails
                    return
                try:
                    set_history_length(DEFAULT_USER_SETTINGS["histsize"])
                except Exception:
                    pass
                atexit.register(self.user_write_history_file)
                pass
        return

    def user_write_history_file(self):
        if self.histfile is not None:
            write_history_file(self.histfile)

    def close(self):
        """Closes both input and output"""
        try:
            self.input.close()
            self.output.close()
        except Exception:
            pass
        return

    def confirm(self, prompt: str, default: bool) -> bool:
        """Called when a dangerous action is about to be done, to make
        sure it's okay. Expect a yes/no answer to `prompt' which is printed,
        suffixed with a question mark and the default value.  The user
        response converted to a boolean is returned."""
        if default:
            prompt += "? (Y or n) "
        else:
            prompt += "? (N or y) "
            pass
        while True:
            try:
                reply = self.readline(prompt)
                reply = reply.strip().lower()
            except EOFError:
                return default
            if reply in ("y", "yes"):
                return True
            elif reply in ("n", "no"):
                return False
            else:
                self.msg("Please answer y or n.")
                pass
            pass
        return default

    def msg(self, msg):
        """used to write to a debugger that is connected to this
        server; `str' written will have a newline added to it
        """
        super().msg(msg)
        # from pydoc import ttypager
        # if hasattr(self.output, "isatty"):
        #     ttypager(msg)
        #     super().msg("")
        # else:
        #     super().msg(msg)

    def errmsg(self, msg, prefix="** "):
        """Common routine for reporting debugger error messages."""
        return self.msg("%s%s" % (prefix, msg))

    def finalize(self, last_wishes=None):
        # This routine gets called multiple times.
        # We hard-code the close() function here.
        try:
            self.msg("%s: That's all, folks..." % self.debugger_name)
        except Exception:
            pass
        else:
            self.close()
            pass
        return

    def read_command(self, prompt=""):
        line = self.readline(prompt)
        # Do something with history?
        return line

    def readline(self, prompt="", add_to_history=True):
        if not self.readline == "prompt_toolkit":
            self.output.flush()
            pass
        return self.input.readline(prompt=prompt)

    def warnmsg(self, msg, prefix="* "):
        """Common routine for reporting debugger warning messages."""
        return self.msg("%s%s" % (prefix, msg))

    pass


# Demo
if __name__ == "__main__":
    print(f"History file is {histfile}")
    intf = UserInterface()
    intf.errmsg("Houston, we have a problem here!")
    import sys

    if len(sys.argv) > 1:
        try:
            line = intf.readline("Type something: ")
        except EOFError:
            print("No input EOF: ")
        else:
            print("You typed: %s" % line)
            pass
        line = intf.confirm("Are you sure", False)
        print("You typed: %s" % line)
        line = intf.confirm("Are you not sure", True)
        print("You typed: %s" % line)
        pass
    pass



# -*- coding: utf-8 -*-

#   Copyright (C) Rocky Bernstein <rocky@gnu.org>
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from prompt_toolkit.completion import CompleteEvent, Completion, WordCompleter
from prompt_toolkit.document import Document
from typing import Iterable, Sequence


class Trepan3KCompleter(WordCompleter):
    def __init__(self, top_level_commands: Sequence[str], aliases):

        # Note: "words" is a parameter WordCompleter uses and sets.
        words = sorted(top_level_commands + list(aliases.keys()))
        super().__init__(words)

        self.states = {"top-level-command": words}
        self.aliases = aliases

    def add_completions(self, key: str, words: Sequence[str]):
        self.states[key] = words

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        # Get list of words.
        text = document.text_before_cursor

        input_words = text.split()
        word_count = len(input_words)
        candidate_completions = []
        if word_count == 0 or (
            word_count == 1
            and not text.endswith(" ")
        ):
            candidate_completions = self.states["top-level-command"]
            if word_count == 1:
                word0 = input_words[0]
                if word0 in self.states["top-level-command"]:
                    display = self.aliases[word0] if word0 in self.aliases else word0
                    display_meta = self.meta_dict.get(display, "")
                    word_before_cursor = document.get_word_before_cursor(
                        WORD=self.WORD, pattern=self.pattern
                    )
                    yield Completion(
                        text=display + " ",
                        start_position=-len(word_before_cursor),
                        display=display,
                        display_meta=display_meta,
                    )
                    return
        elif word_count >= 1:
            if not text.endswith(" "):
                # Matching with a partially-filled final word
                input_words = input_words[:-1]

            key = ""
            for input_word in input_words:
                key += input_word
                next_state = self.states.get(key)
                if next_state is None:
                    return
                key += " "
            candidate_completions = self.states.get(key.strip())
            if candidate_completions is None:
                return

        # Get word/text before cursor.
        word_before_cursor = document.get_word_before_cursor(
            WORD=self.WORD, pattern=self.pattern
        )

        def word_matches(word: str) -> bool:
            """True when the word before the cursor matches."""
            return word.startswith(word_before_cursor)

        for candidate_word in candidate_completions:
            if word_matches(candidate_word):
                display = self.display_dict.get(candidate_word, candidate_word)
                display_meta = self.meta_dict.get(candidate_word, "")
                yield Completion(
                    text=candidate_word,
                    start_position=-len(word_before_cursor),
                    display=display,
                    display_meta=display_meta,
                )


if __name__ == "__main__":
    from prompt_toolkit import PromptSession
    from trepan import debugger

    d = debugger.Trepan()
    proc = d.core.processor

    command_names = list(proc.commands.keys())
    trepan3k_completer = Trepan3KCompleter(command_names, proc.aliases)
    for cmd, cmd_obj in proc.commands.items():
        cmd_obj = proc.commands[cmd]
        if hasattr(cmd_obj, "cmds") and hasattr(cmd_obj.cmds, "cmdlist"):
            trepan3k_completer.add_completions(cmd, cmd_obj.cmds.cmdlist)
            for subcmd_name, subcmd_obj in cmd_obj.cmds.subcmds.items():
                subcmd_key = f"{cmd} {subcmd_name}"
                if hasattr(subcmd_obj, "completion_choices"):
                    trepan3k_completer.add_completions(subcmd_key, subcmd_obj.completion_choices)
                pass
            pass
        elif hasattr(cmd_obj, "completion_choices"):
            trepan3k_completer.add_completions(cmd, cmd_obj.completion_choices)

        pass

    def main():
        session = PromptSession(completer=trepan3k_completer)

        while True:
            try:
                text = session.prompt("> ")
            except KeyboardInterrupt:
                continue
            except EOFError:
                break
            else:
                print("You entered:", text)
        print("GoodBye!")

    main()



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2020, 2024 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class StepICommand(DebuggerCommand):
    """**stepi** [*count*]

    Execute the current line, stopping at the instruction bytecode.

    With an integer argument, step bytecode instructions that many times.


    Examples:
    ---------

      stepi        # step 1 bytecode instruction
      si 1         # same as above

    Related and similar is the `step` command.

    See also:
    ---------

    `step`, `next`, `skip`, `jump` (there's no `hop` yet), `continue`, `return` and
    `finish` for other ways to progress execution."""

    aliases = ("si",)
    execution_set = ["Running"]
    short_help = "Step bytecode instruction (possibly entering called functions)"

    DebuggerCommand.setup(locals(), category="running", max_args=1, need_stack=True)

    def run(self, args):
        if len(args) <= 1:
            self.proc.debugger.core.step_ignore = 0
        else:
            pos = 1
            if pos == len(args) - 1:
                self.core.step_ignore = self.proc.get_int(
                    args[pos], default=1, cmdname="stepi"
                )
                if self.core.step_ignore is None:
                    return False
                # 0 means stop now or step 1, so we subtract 1.
                self.core.step_ignore -= 1
                pass
            elif pos != len(args):
                self.errmsg(f"Invalid additional parameters {' '.join(args[pos])}")
                return False
            pass

        self.core.step_events = None

        self.core.different_line = False
        # print("XXX", self.proc.frame)
        if self.proc.frame is not None:
            self.proc.frame.f_trace_opcodes = True
        self.core.stop_level = None
        self.core.last_frame = None
        self.core.stop_on_finish = False
        self.proc.continue_running = True  # Break out of command read loop
        return True

    pass


if __name__ == "__main__":
    from mock import MockDebugger

    d = MockDebugger()
    cmd = StepICommand(d.core.processor)
    for c in (["si", "5"], ["stepi", "1+2"], ["si", "foo"]):
        d.core.step_ignore = 0
        cmd.proc.continue_running = False
        result = cmd.run(c)
        print(f"Execute result: {result}")
        print(f"step_ignore {repr(d.core.step_ignore)}")
        print(f"continue_running: {cmd.proc.continue_running}")
        pass
    for c in (["si"], ["stepi"]):
        d.core.step_ignore = 0
        cmd.continue_running = False
        result = cmd.run(c)
        print(f"different line {c[0]}:", cmd.core.different_line)
        pass
    pass



# -*- coding: utf-8 -*-
# Copyright (C) 2024 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import inspect

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.command.base_submgr import SubcommandMgr


class ReloadCommand(DebuggerCommand):
    """**reload** *command-name*
       **reload** *subcommand-name* *subcommand*

    Reload debugger command *command-name* or subcommand *subcommand-name* *subcommand*.

    This command is usually used by trepan3k developers working on the program

    Examples:
    --------

        reload alias
        reload show listsize
    """

    short_help = "reload a trepan3k command"

    DebuggerCommand.setup(locals(), category="support", min_args=1, max_args=2)

    def run(self, args):
        cmd_name = args[1]
        proc = self.proc
        if len(args) == 2:
            cmd_name = proc.aliases.get(cmd_name, cmd_name)
            if cmd_name not in proc.commands:
                self.errmsg(f'command "{cmd_name}" not found as a debugger command')
                return
            command_module = importlib.import_module(proc.commands[cmd_name].__module__)
            importlib.reload(command_module)
            classnames = [
                tup[0]
                for tup in inspect.getmembers(command_module, inspect.isclass)
                if ("DebuggerCommand" != tup[0] and tup[0].endswith("Command"))
            ]
            if len(classnames) == 1:
                try:
                    instance = getattr(command_module, classnames[0])(proc)
                except Exception:
                    self.errmsg(
                        f"Error loading {classnames[0]} from module name, sys.exc_info()[0]"
                    )
                    return

                # FIXME: should we also replace object in proc.cmd_instances?
                proc.commands[cmd_name] = instance
                self.msg(f'reloaded command: "{cmd_name}"')
            pass
        else:
            assert len(args) == 3
            subcmd_mgr = proc.commands.get(cmd_name)
            if subcmd_mgr is None:
                self.errmsg(f"cannot find {cmd_name} in list of commands")
                return
            if not isinstance(subcmd_mgr, SubcommandMgr):
                self.errmsg(f"command {cmd_name} does not have subcommands")
                return
            subcmd_name = args[2]
            subcmd = subcmd_mgr.cmds.subcmds.get(subcmd_name)
            if subcmd is None:
                self.errmsg(
                    f'command "{cmd_name}" does not have subcommand ' f'"{subcmd_name}"'
                )
                return

            subcommand_module = importlib.import_module(subcmd.__module__)
            importlib.reload(subcommand_module)
            classnames = [
                tup[0] for tup in inspect.getmembers(subcommand_module, inspect.isclass)
            ]
            if len(classnames) == 1:
                try:
                    instance = getattr(subcommand_module, classnames[0])(subcmd)
                except Exception:
                    self.errmsg(
                        f"Error loading {classnames[0]} from mod_name, sys.exc_info()[0]"
                    )
                    return

                subcmd_mgr.cmds.subcmds[subcmd_name] = instance
                self.msg(f'reloaded subcommand: "{cmd_name} {subcmd_name}"')
            return

    pass


# Demo it
if __name__ == "__main__":
    from trepan.processor.command import mock as Mmock

    dbgr, cmd = Mmock.dbg_setup()
    command = ReloadCommand(cmd)
    for cmdline in [
        "reload up",
        "reload show listsize",
        "reload foo",
        "reload show foo",
    ]:
        args = cmdline.split()
        cmd_argstr = cmdline[len(args[0]) :].lstrip()
        cmd.cmd_argstr = cmd_argstr
        command.run(args)
        pass
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2024 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.lib.complete import complete_token
from trepan.processor.command.base_subcmd import DebuggerSubcommand

from trepan.interfaces.user import DEFAULT_USER_SETTINGS

class ShowHistory(DebuggerSubcommand):
    """**show history** [ *filename* | *save* | *size* ]

    Show history information

    options include:
    * size: Show the size of the command history.
    * filename Show the filename in which to record the command historyh.

    Examples:
    --------
        (trepan3k) show history size
        The size of the command history is 50.

        (trepan3k) show history
        history filename: The filename in which to record the command history is "/home/joe/.trepan3k_hist".
        history save: Saving of the history record on exit is on.
        history size: The size of the command history is 50.
    """

    min_abbrev = len("his")
    need_stack = False
    short_help = "Generic command for showing command history parameters"

    def complete(self, prefix):
        return complete_token(("filename", "save", "size"))

    def run(self, args):
        opts = self.debugger.settings
        show_prefix = False
        if len(args) == 0:
            args = ("filename", "save", "size")
            show_prefix = True

        for arg in args:
            prefix = f"history {arg}: " if show_prefix else ""
            if arg == "filename":
                self.msg(f'{prefix}The filename in which to record the command history is "{self.proc.intf[-1].histfile}".')
            elif arg == "save":
                self.msg(f'{prefix}Saving of the history record on exit is {"on" if opts.get("hist_save", False) else "off"}.')
            elif arg == "size":
                self.msg(f'{prefix}The size of the command history is {DEFAULT_USER_SETTINGS["histsize"]}.')
            else:
                self.errmsg(f"Undefined show history command: {arg}")
                pass
        return

    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd.__demo_helper__ import demo_run

    sub = demo_run(ShowHistory, [])
    d = sub.proc.debugger
    sub.run([])
    sub.run(["filename"])
    sub.run(["save"])
    sub.run(["size"])
    sub.run(["foo"])
    pass



# Copied from 3.11
"""Utilities for with-statement contexts.  See PEP 343."""

class ChDir(object):
    """
    Step into a directory temporarily.
    """
    def __init__(self, path):
        self.old_dir = os.getcwd()
        self.new_dir = path

    def __enter__(self):
        os.chdir(self.new_dir)

    def __exit__(self, *args):
        os.chdir(self.old_dir)



# -*- coding: utf-8 -*-
#  Copyright (C) 2009-2013, 2015-2016, 2020, 2024 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.complete_rl import complete_id_and_builtins


class DisplayCommand(DebuggerCommand):
    """**display** [*format*] *expression*

    Print value of expression *expression* each time the program stops.
    *format* may be used before *expression* and may be one of `/c` for
    char, `/x` for hex, `/o` for octal, `/f` for float or `/s` for string.

    For now, display expressions are only evaluated when in the same
    code as the frame that was in effect when the display expression
    was set.  This is a departure from gdb and we may allow for more
    flexibility in the future to specify whether this should be the
    case or not.

    With no argument, evaluate and display all currently requested
    auto-display expressions.  Use `undisplay` to cancel display
    requests previously made."""

    short_help = "Display expressions when entering debugger"

    format_specs = ("/c", "/x", "/o", "/f", "/s")

    DebuggerCommand.setup(locals(), category="data")

    def complete(self, prefix):
        return DisplayCommand.format_specs + complete_id_and_builtins(self, prefix)

    def run_eval_display(self, args=None):
        for line in self.proc.display_mgr.display(self.proc.curframe):
            self.msg(line)
        return

    def run(self, args):
        if len(args) == 1:
            # Display anything active
            self.run_eval_display(self)
        else:
            if args[1] in DisplayCommand.format_specs:
                if len(args) == 2:
                    self.errmsg("Expecting an expression after the format")
                    return
                format = args[1]
                expr = " ".join(args[2:])
            else:
                format = None
                expr = " ".join(args[1:])
                pass
            dp = self.proc.display_mgr.add(self.proc.curframe, expr, format)
            if dp is None:
                self.errmsg('Error evaluating "%s" in the current frame' % expr)
                return
            self.msg(dp.format(show_enabled=False))
            self.proc.add_preloop_hook(self.run_eval_display)
            self.msg(dp.to_s(self.proc.curframe))
        return False

    pass


if __name__ == "__main__":
    from trepan.processor import cmdproc as Mcmdproc
    from trepan.debugger import Trepan

    d = Trepan()
    import inspect

    d = Trepan()
    cp = d.core.processor
    command = DisplayCommand(d.core.processor)
    cp.curframe = inspect.currentframe()
    cp.stack, cp.curindex = Mcmdproc.get_stack(cp.curframe, None, None, cp)
    command.run(["display"])
    command.run(["display", "/x", "10"])
    command.run(["display", "d"])
    print("==" * 10)
    command.run(["display"])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2015, 2023-2024 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect

from pyficache import getline, highlight_string

from trepan.lib.complete import complete_token
from trepan.lib.disassemble import opc
from trepan.lib.format import Function, Keyword, format_token
from trepan.lib.stack import format_function_name
from trepan.lib.thred import current_thread_name
from trepan.processor import frame as Mframe
from trepan.processor.print import format_code, format_frame

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class InfoFrame(Mbase_subcmd.DebuggerSubcommand):
    """**info frame** [-v] [ *frame-number* | *frame-object* ]

    Show the detailed information for *frame-number* or the current frame if
    *frame-number* is not specified. You can also give a frame object instead of
    a frame number

    Specific information includes:

    * the frame number (if not an object)

    * the source-code line number that this frame is stopped in

    * the last instruction executed; -1 if the program are before the first
    instruction

    * a function that tracing this frame or `None`

    * Whether the frame is in restricted execution

    If `-v` is given we show builtin and global names the frame sees.

    See also:
    ---------

    `info locals`, `info globals`, `info args`"""

    min_abbrev = 2
    max_args = 3
    need_stack = True
    short_help = """Show detailed info about the current frame"""

    def complete(self, prefix):
        proc_obj = self.proc
        low, high = Mframe.frame_low_high(proc_obj, None)
        ary = [str(low + i) for i in range(high - low + 1)]
        # FIXME: add in Thread names
        return complete_token(ary, prefix)

    def run(self, args):
        # FIXME: should DRY this with code.py
        proc = self.proc
        frame = proc.curframe
        if not frame:
            self.errmsg("No frame selected.")
            return False

        is_verbose = False
        if len(args) >= 1 and args[0] == "-v":
            args.pop(0)
            is_verbose = True

        style = self.settings["style"]
        frame_num = None
        if len(args) == 1:
            try:
                frame_num = int(args[0])
                i_stack = len(proc.stack)
                if frame_num < -i_stack or frame_num > i_stack - 1:
                    self.errmsg(
                        (
                            "a frame number number has to be in the range %d to %d."
                            + " Got: %d (%s)."
                        )
                        % (-i_stack, i_stack - 1, frame_num, args[0])
                    )
                    return False
                frame = proc.stack[frame_num][0]
            except Exception:
                try:
                    frame = eval(args[0], frame.f_globals, frame.f_locals)
                except Exception:
                    self.errmsg(
                        "%s is not a evaluable as a frame object or frame number."
                        % args[0]
                    )
                    return False
                if not inspect.isframe(frame):
                    self.errmsg(f"{frame} is not a frame object")
                pass
        else:
            frame_num = proc.curindex

        mess = (
            f"Frame {Mframe.frame_num(proc, frame_num)}"
            if frame_num is not None and proc.stack is not None
            else "Frame Info"
        )
        self.section(mess)

        function_name, formatted_func_name = format_function_name(frame, style=style)
        if function_name is None:
            formatted_func_name = "??"
        f_args, f_varargs, f_keywords, f_locals = inspect.getargvalues(frame)
        self.msg(f"  function name: {formatted_func_name}")
        func_args = inspect.formatargvalues(f_args, f_varargs, f_keywords, f_locals)
        formatted_func_signature = highlight_string(func_args, style=style).strip()
        self.msg(f"  function args: {formatted_func_signature}")

        formatted_thread_name = format_token(
            Function, current_thread_name(), style=style
        )
        self.msg(f"  thread: {formatted_thread_name}")
        # signature = highlight_string(inspect.signature(frame))
        # self.msg(f"  signature : {signature}")

        if hasattr(frame, "f_restricted"):
            self.msg(f"  restricted execution: {frame.f_restricted}")

        line_number = frame.f_lineno
        code = frame.f_code
        file_path = code.co_filename

        line_text = getline(file_path, line_number, {"style": style})
        if line_text is None:
            self.msg(f"  current line number: {frame.f_lineno}")
        else:
            formatted_text = highlight_string(line_text.strip(), style=style)
            self.msg_nocr(f"  current line number: {frame.f_lineno}: {formatted_text}")

        f_lasti = frame.f_lasti
        if f_lasti >= 0:
            opname = opc.opname[code.co_code[f_lasti]]
            opname_formatted = format_token(Keyword, opname, style=style)
            self.msg(f"  instruction offset and opname: {frame.f_lasti} {opname_formatted}")
        else:
            self.msg(f"  instruction offset: {frame.f_lasti}")

        self.msg(f"  code: {format_code(code, style)}")
        if frame.f_back:
            self.msg(f"  previous frame: {format_frame(frame.f_back, style)}")
        else:
            self.msg("  no previous frame")
        self.msg(f"  tracing function: {frame.f_trace}")
        if hasattr(frame, "f_trace_opcodes"):
            self.msg_nocr(
                f"  tracing opcodes: {highlight_string(str(frame.f_trace_opcodes), style=style)}"
            )
        if hasattr(frame, "f_trace_lines"):
            self.msg(
                f"  tracing lines: {highlight_string(str(frame.f_trace_lines), style=style)}"
            )

        if is_verbose:
            for name, field in [
                ("Globals", "f_globals"),
                ("Builtins", "f_builtins"),
            ]:
                # FIXME: not sure this is quite right.
                # For now we'll strip out values that start with the option
                # prefix "-".
                vals = [
                    field
                    for field in getattr(frame, field).keys()
                    if not field.startswith("-")
                ]
                if vals:
                    self.section(name)
                    m = self.columnize_commands(vals)
                    self.msg_nocr(m)

        return False

    pass


if __name__ == "__main__":
    from trepan.processor.command import info as Minfo, mock

    d, cp = mock.dbg_setup()
    cp.setup()
    i = Minfo.InfoCommand(cp)

    sub = InfoFrame(i)
    cp.curframe = inspect.currentframe()
    sub.run([])



import difflib
import os
import os.path as osp
import re
import sys
import time

from xdis import IS_PYPY

srcdir = osp.abspath(osp.dirname(__file__))


def run_debugger(
    testname, python_file, dbgr_opts="", args="", outfile=None, right_template=None
):
    datadir = osp.join(srcdir, "..", "data")
    progdir = osp.join(srcdir, "..", "example")
    dbgrdir = osp.join(srcdir, "..", "..", "trepan")
    dbgr_short = "__main__.py"
    dbgr_path = osp.join(dbgrdir, dbgr_short)

    if not right_template:
        if IS_PYPY:
            right_template = "%s-pypy.right"
        else:
            right_template = "%s.right"

    rightfile = osp.join(datadir, right_template % testname)

    sys.path.insert(0, osp.join(srcdir, "..", ".."))
    os.environ["PYTHONPATH"] = os.pathsep.join(sys.path)
    cmdfile = osp.join(datadir, f"{testname}.cmd")
    outfile = osp.join(srcdir, f"{testname}.out")
    if python_file:
        programfile = osp.join(progdir, python_file)
    else:
        programfile = ""
        pass

    outfile_opt = f"--output={outfile} "

    if osp.exists(outfile):
        os.unlink(outfile)

    cmd = "%s --command %s %s %s %s %s" % (
        dbgr_path,
        cmdfile,
        outfile_opt,
        dbgr_opts,
        programfile,
        args,
    )

    # print(cmd)
    os.system(cmd)
    fromfile = rightfile
    fromdate = time.ctime(os.stat(fromfile).st_mtime)
    tofile = outfile
    todate = time.ctime(os.stat(tofile).st_mtime)
    with open(fromfile) as f:
        fromlines = f.readlines()
    with open(tofile) as f:
        tolines = f.readlines()

    # Filter out last instruction. For example:
    # (gcd.py:11 @6): -> (gcd.py:11)
    tolines = [re.sub(r" @\d+\):", "):", line) for line in tolines]

    diff = list(
        difflib.unified_diff(fromlines, tolines, fromfile, tofile, fromdate, todate)
    )
    if len(diff) == 0:
        os.unlink(outfile)
        pass
    else:
        with open(tofile, "w") as out:
            out.writelines(tolines)
            pass
        pass
    for line in diff:
        print(line.rstrip())
        pass
    return len(diff) == 0
